import io
import os
import sys

import click
import resumable

from . import client as askanna_client
from askanna.core.utils import _file_type, diskunit


class Upload:
    tpl_register_upload_url = "{ASKANNA_API_SERVER}package/"
    tpl_register_chunk_url = "{ASKANNA_API_SERVER}package/{PACKAGE_SUUID}/packagechunk/"
    tpl_upload_chunk_url = (
        "{ASKANNA_API_SERVER}package/{PACKAGE_SUUID}/packagechunk/{CHUNK_UUID}/chunk/"
    )
    tpl_final_upload_url = "{ASKANNA_API_SERVER}package/{PACKAGE_SUUID}/finish_upload/"

    tpl_upload_pass = "File is uploaded"
    tpl_upload_fail = "Upload failed"

    def __init__(self, api_server: str = None, *args, **kwargs):
        self.ASKANNA_API_SERVER = api_server or askanna_client.config.remote
        self.suuid = None
        self.resumable_file = None
        self.kwargs = kwargs

    @property
    def chunk_baseinfo(self) -> dict:
        return {
            "resumableChunkSize": self.resumable_file.chunk_size,
            "resumableTotalSize": self.resumable_file.size,
            "resumableType": _file_type(self.resumable_file.path),
            "resumableIdentifier": str(self.resumable_file.unique_identifier),
            "resumableFilename": os.path.basename(self.resumable_file.path),
            "resumableRelativePath": self.resumable_file.path,
            "resumableTotalChunks": len(self.resumable_file.chunks),
            "resumableChunkNumber": 1,
            "resumableCurrentChunkSize": 1,
        }.copy()

    def url_template_arguments(self):
        """
        This method is meant to be overwritten by inheritors
        providing specific url templating variables
        """
        return {
            "ASKANNA_API_SERVER": self.ASKANNA_API_SERVER,
            "PACKAGE_SUUID": self.suuid,
        }

    @property
    def register_upload_url(self) -> str:
        return self.tpl_register_upload_url.format(**self.url_template_arguments())

    @property
    def register_chunk_url(self) -> str:
        return self.tpl_register_chunk_url.format(**self.url_template_arguments())

    def upload_chunk_url(self, chunk_uuid: str) -> str:
        arguments = self.url_template_arguments()
        arguments.update({"CHUNK_UUID": chunk_uuid})
        return self.tpl_upload_chunk_url.format(**arguments)

    @property
    def final_upload_url(self) -> str:
        return self.tpl_final_upload_url.format(**self.url_template_arguments())

    def upload(self, file_obj, config, fileinfo):
        self.create_entry(config=config, fileinfo=fileinfo)
        self.upload_file(file_obj)
        return self.finish_upload()

    def create_entry_extrafields(self):
        return {}

    def create_entry(self, config: dict = None, fileinfo: dict = {}) -> str:
        info_dict = {
            "filename": fileinfo.get("filename"),
            "size": fileinfo.get("size"),
        }

        info_dict.update(**self.create_entry_extrafields())

        # the request to AskAnna API
        req = askanna_client.post(self.register_upload_url, json=info_dict)
        if req.status_code != 201:
            click.echo(
                "In the AskAnna platform something went wrong with creating the upload entry.",
                err=True,
            )
            sys.exit(1)

        res = req.json()

        # the result
        self.suuid = res.get("short_uuid")
        return self.suuid

    def upload_chunk(self, chunk, chunk_dict):
        config = chunk_dict.copy()
        config.update(
            **{
                "filename": chunk.index + 1,
                "size": chunk.size,
                "file_no": chunk.index + 1,
                "is_last": len(self.resumable_file.chunks) == chunk.index + 1,
            }
        )

        # request chunk id from API
        req_chunk = askanna_client.post(self.register_chunk_url, json=config)
        chunk_uuid = req_chunk.json().get("uuid")

        files = {"file": io.BytesIO(chunk.read())}
        data = self.chunk_baseinfo
        data.update(
            **{
                "resumableChunkNumber": chunk.index + 1,
                "resumableCurrentChunkSize": chunk.size,
            }
        )

        specific_chunk_req = askanna_client.post(
            self.upload_chunk_url(chunk_uuid=chunk_uuid), data=data, files=files
        )
        assert specific_chunk_req.status_code == 200, "File could not be uploaded"

    def chunk_dict_template(self):
        return {
            "filename": "",
            "size": 0,
            "file_no": 0,
            "is_last": False,
        }

    def upload_file(self, file_obj):
        """
        Take a file_obj and make chunks out of it to upload
        """
        chunk_dict = self.chunk_dict_template()

        self.resumable_file = resumable.file.ResumableFile(file_obj, 1 * diskunit.MiB)
        for chunk in self.resumable_file.chunks:
            self.upload_chunk(chunk, chunk_dict)

    def finish_upload(self):
        # Do final call when all chunks are uploaded
        final_call_dict = self.chunk_baseinfo

        final_call_req = askanna_client.post(
            self.final_upload_url, data=final_call_dict
        )

        if final_call_req.status_code == 200:
            return True, self.tpl_upload_pass
        else:
            return False, self.tpl_upload_fail


class PackageUpload(Upload):
    tpl_register_upload_url = "{ASKANNA_API_SERVER}package/"
    tpl_register_chunk_url = "{ASKANNA_API_SERVER}package/{PACKAGE_SUUID}/packagechunk/"
    tpl_upload_chunk_url = (
        "{ASKANNA_API_SERVER}package/{PACKAGE_SUUID}/packagechunk/{CHUNK_UUID}/chunk/"
    )
    tpl_final_upload_url = "{ASKANNA_API_SERVER}package/{PACKAGE_SUUID}/finish_upload/"

    tpl_upload_pass = "Package is uploaded"
    tpl_upload_fail = "Package upload failed"

    def create_entry_extrafields(self):
        return {
            "project": self.kwargs.get("project_suuid"),
            "description": self.kwargs.get("description"),
        }


class ArtifactUpload(Upload):
    tpl_register_upload_url = "{ASKANNA_API_SERVER}runinfo/{RUN_SUUID}/artifact/"
    tpl_register_chunk_url = "{ASKANNA_API_SERVER}runinfo/{RUN_SUUID}/artifact/{ARTIFACT_SUUID}/artifactchunk/"
    tpl_upload_chunk_url = "{ASKANNA_API_SERVER}runinfo/{RUN_SUUID}/artifact/{ARTIFACT_SUUID}/artifactchunk/{CHUNK_UUID}/chunk/"  # noqa
    tpl_final_upload_url = "{ASKANNA_API_SERVER}runinfo/{RUN_SUUID}/artifact/{ARTIFACT_SUUID}/finish_upload/"

    tpl_upload_pass = "Artifact is uploaded"
    tpl_upload_fail = "Artifact upload failed"

    def url_template_arguments(self):
        """
        This method is meant to be overwritten by inheritors
        providing specific url templating variables
        """
        return {
            "ASKANNA_API_SERVER": self.ASKANNA_API_SERVER,
            "ARTIFACT_SUUID": self.suuid,
            "RUN_SUUID": self.kwargs.get("run_suuid"),
        }


class ResultUpload(Upload):
    tpl_register_upload_url = "{ASKANNA_API_SERVER}runinfo/{RUN_SUUID}/result/"
    tpl_register_chunk_url = (
        "{ASKANNA_API_SERVER}runinfo/{RUN_SUUID}/result/{RESULT_SUUID}/resultchunk/"
    )
    tpl_upload_chunk_url = "{ASKANNA_API_SERVER}runinfo/{RUN_SUUID}/result/{RESULT_SUUID}/resultchunk/{CHUNK_UUID}/chunk/"  # noqa
    tpl_final_upload_url = (
        "{ASKANNA_API_SERVER}runinfo/{RUN_SUUID}/result/{RESULT_SUUID}/finish_upload/"
    )

    tpl_upload_pass = "Result is uploaded"
    tpl_upload_fail = "Result upload failed"

    def url_template_arguments(self):
        return {
            "ASKANNA_API_SERVER": self.ASKANNA_API_SERVER,
            "RUN_SUUID": self.kwargs.get("RUN_SUUID"),
            "RESULT_SUUID": self.suuid,
        }



# -*- coding: utf-8 -*-
import os
import sys
import click

from askanna.core.utils import get_config
from askanna.core.upload import ResultUpload

HELP = """
At the end of a run push the result to AskAnna
"""

SHORT_HELP = "Push result to AskAnna"


@click.command(help=HELP, short_help=SHORT_HELP)
def cli():
    config = get_config()

    run_suuid = os.getenv("AA_RUN_SUUID")
    job_name = os.getenv("AA_JOB_NAME")

    # First check whether we need to create result or not.
    # If output.result is not specified, we skip this step and report this to the stdout.
    result_path = config[job_name].get("output", {}).get("result")

    if not result_path:
        click.echo(
            "Result storage aborted. No `output/result` defined for this job in `askanna.yml`."
        )
        sys.exit(0)
    elif isinstance(result_path, list):
        click.echo("Please enter a path in `output/result`, not a list.", err=True)
        sys.exit(1)
    elif not os.path.exists(result_path):
        click.echo(
            f"output/result: {result_path} does not exist. Not saving result.", err=True
        )
        sys.exit(1)

    click.echo("Uploading result to AskAnna...")

    fileinfo = {
        "filename": os.path.basename(result_path),
        "size": os.stat(result_path).st_size,
    }
    uploader = ResultUpload(RUN_SUUID=run_suuid)
    status, msg = uploader.upload(result_path, config, fileinfo)
    if status:
        click.echo(msg)
        sys.exit(0)
    else:
        click.echo(msg, err=True)
        sys.exit(1)



from click.testing import CliRunner
import os
import responses
import shutil
import tempfile
import unittest

from askanna.cli.tool import cli_commands
from askanna.core import client
from tests.create_fake_files import create_json_file


class TestCLIResult(unittest.TestCase):
    """
    askanna result

    We expect to be able to get result from a run
    """

    def setUp(self):
        self.environ_bck = dict(os.environ)
        os.environ["AA_REMOTE"] = "https://beta-api.askanna.eu/v1/"
        os.environ["AA_TOKEN"] = "12345678910"

        self.tempdir = tempfile.mkdtemp(prefix="askanna-test-cli-result")
        self.result_json_file = create_json_file(self.tempdir, 10)

        with open(self.result_json_file, "rb") as f:
            content = f.read()

        self.base_url = client.config.remote
        self.responses = responses.RequestsMock()
        self.responses.start()
        self.responses.add(
            responses.HEAD,
            url=self.base_url + "result/abcd-abcd-abcd-abcd/",
            headers={'Content-Length': str(os.path.getsize(self.result_json_file)), 'Accept-Ranges': 'bytes'},
            content_type="application/json",
            status=200,
        )
        self.responses.add(
            responses.HEAD,
            url=self.base_url + "result/1234-1234-1234-1234/",
            status=404,
        )
        self.responses.add(
            responses.GET,
            url=self.base_url + "result/abcd-abcd-abcd-abcd/",
            headers={"Range": f"bytes=0-{os.path.getsize(self.result_json_file)}"},
            stream=True,
            content_type="application/json",
            status=206,
            body=content
        )

    def tearDown(self):
        self.responses.stop
        self.responses.reset

        os.remove(self.result_json_file)
        shutil.rmtree(self.tempdir, ignore_errors=True)

        os.environ.clear()
        os.environ.update(self.environ_bck)

    def test_command_result_base(self):
        result = CliRunner().invoke(cli_commands, "result --help")

        assert not result.exception
        assert "result" in result.output
        assert "noop" not in result.output

    def test_command_result_get(self):
        result = CliRunner().invoke(cli_commands, "result get --help")

        assert "result" in result.output
        assert "get [OPTIONS]" in result.output
        assert "noop" not in result.output

    def test_command_result_get_prompt_invalid_suuid(self):
        result = CliRunner().invoke(cli_commands, "result get", input="test")

        assert result.exception
        assert "Run SUUID: test" in result.output

    def test_command_result_get_unknown_file(self):
        result = CliRunner().invoke(cli_commands, "result get", input="1234-1234-1234-1234")

        assert result.exception
        assert "Run SUUID: 1234-1234-1234-1234" in result.output
        assert "404 - We cannot find this result for you" in result.output

    def test_command_result_get_prompt(self):
        result = CliRunner().invoke(cli_commands, "result get", input="abcd-abcd-abcd-abcd")
        os.remove("result_abcd-abcd-abcd-abcd.json")

        assert not result.exception
        assert "Run SUUID: abcd-abcd-abcd-abcd" in result.output
        assert "Downloading the result has started..." in result.output
        assert "The result is saved in: result_abcd-abcd-abcd-abcd.json" in result.output

    def test_command_result_get_already_exist(self):
        CliRunner().invoke(cli_commands, "result get", input="abcd-abcd-abcd-abcd")
        result = CliRunner().invoke(cli_commands, "result get", input="abcd-abcd-abcd-abcd")
        os.remove("result_abcd-abcd-abcd-abcd.json")

        assert result.exception
        assert "Run SUUID: abcd-abcd-abcd-abcd" in result.output
        assert "The output file already exists. We will not overwrite the existing file." in result.output

    def test_command_result_get_argument_full(self):
        result = CliRunner().invoke(cli_commands, "result get --id abcd-abcd-abcd-abcd")
        os.remove("result_abcd-abcd-abcd-abcd.json")

        assert not result.exception
        assert "Downloading the result has started..." in result.output
        assert "The result is saved in: result_abcd-abcd-abcd-abcd.json" in result.output

    def test_command_result_get_argument_short(self):
        result = CliRunner().invoke(cli_commands, "result get -i abcd-abcd-abcd-abcd")
        os.remove("result_abcd-abcd-abcd-abcd.json")

        assert not result.exception
        assert "Downloading the result has started..." in result.output
        assert "The result is saved in: result_abcd-abcd-abcd-abcd.json" in result.output

    def test_command_result_get_argument_output_dir(self):
        result = CliRunner().invoke(cli_commands, f"result get --id abcd-abcd-abcd-abcd --output {self.tempdir}")

        assert result.exception
        assert "The output argument is a directory. Please provide a file name for the output." in result.output

    def test_command_result_get_argument_output_file(self):
        output_file = self.tempdir + "/result-json-abcd-abcd-abcd-abcd.json"
        result = CliRunner().invoke(cli_commands, f"result get --id abcd-abcd-abcd-abcd --output {output_file}")

        self.assertEqual(os.path.getsize(self.result_json_file), os.path.getsize(output_file))
        os.remove(output_file)

        assert not result.exception
        assert "Downloading the result has started..." in result.output
        assert f"The result is saved in: {output_file}" in result.output

    def test_command_result_get_argument_output_file_short(self):
        output_file = self.tempdir + "/result-json-abcd-abcd-abcd-abcd.json"
        result = CliRunner().invoke(cli_commands, f"result get --id abcd-abcd-abcd-abcd -o {output_file}")

        self.assertEqual(os.path.getsize(self.result_json_file), os.path.getsize(output_file))
        os.remove(output_file)

        assert not result.exception
        assert "Downloading the result has started..." in result.output
        assert f"The result is saved in: {output_file}" in result.output



# -*- coding: utf-8 -*-
import json
import sys

import click

from askanna import config
from askanna import run as askanna_run
from askanna import job as askanna_job
from askanna.core.dataclasses import Project
from askanna.cli.utils import ask_which_job, ask_which_project, ask_which_workspace
from askanna.core.push import push
from askanna.core.utils import getProjectInfo


HELP = """
This command will allow you to start a run in AskAnna.
"""

SHORT_HELP = "Start a run in AskAnna"


def determine_project(
    project_suuid: str = None,
    workspace_suuid: str = None,
) -> Project:
    if not project_suuid:
        project_suuid = config.project_suuid

    # Still if there is no project_suuid found, we will ask which project to use
    if project_suuid:
        project = getProjectInfo(project_suuid=project_suuid)
        click.echo(f"Selected project: {project.name}")

        return project
    else:
        if not workspace_suuid:
            workspace = ask_which_workspace(
                question="From which workspace do you want to run a job?"
            )
            workspace_suuid = workspace.short_uuid

        return ask_which_project(
            question="From which project do you want to run a job?",
            workspace_suuid=workspace_suuid,
        )


@click.command(help=HELP, short_help=SHORT_HELP)
@click.argument(
    "job_name",
    required=False,
    type=str,
)
@click.option("--id", "-i", "job_suuid", required=False, type=str, help="Job SUUID")
@click.option("--data", "-d", required=False, type=str, default=None, help="JSON data")
@click.option(
    "--data-file",
    "-D",
    "data_file",
    required=False,
    type=str,
    default=None,
    help="File with JSON data",
)
@click.option(
    "--push/--no-push",
    "-p",
    "push_code",
    default=False,
    show_default=False,
    help="Push code first, and then run the job [default: no-push]",
)
@click.option("--name", "-n", required=False, type=str, help="Give the run a name")
@click.option(
    "--description",
    required=False,
    type=str,
    help="Description of the run",
    default="",
)
@click.option(
    "--message",
    "-m",
    required=False,
    type=str,
    help="[deprecated] Description of the run",
    default="",
)
@click.option(
    "--project", "project_suuid", required=False, type=str, help="Project SUUID"
)
@click.option(
    "--workspace", "workspace_suuid", required=False, type=str, help="Workspace SUUID"
)
def cli(
    job_name,
    job_suuid,
    name,
    description,
    message,
    data,
    data_file,
    project_suuid,
    workspace_suuid,
    push_code,
):
    if len(description) > 0 and len(message) > 0:
        click.echo(
            "Warning: usage of --message is deprecated. Please use --description."
        )
        click.echo("Cannot use both --description and --message.", err=True)
        sys.exit(1)
    elif len(message) > 0:
        click.echo(
            "Warning: usage of --message is deprecated. Please use --description."
        )
        description = message

    if data and data_file:
        click.echo("Cannot use both --data and --data-file.", err=True)
        sys.exit(1)
    elif data:
        data = json.loads(data)
    elif data_file:
        with open(data_file) as json_file:
            data = json.load(json_file)

    if push_code:
        push(force=True, description=description)

    # Only determine project when it's necessary
    project = None
    if not job_suuid:
        project = determine_project(project_suuid, workspace_suuid)

    if job_suuid:
        pass
    elif job_name:
        try:
            job = askanna_job.get_job_by_name(
                job_name=job_name, project_suuid=project.short_uuid
            )
            job_suuid = job.short_uuid
        except Exception as e:
            click.echo(e)
            sys.exit(1)
    else:
        job = ask_which_job(
            question="Which job do you want to run?",
            project_suuid=project.short_uuid,
        )

        if not click.confirm(f"\nDo you want to run the job '{job.name}'?", abort=True):
            click.echo(f"Aborted! Not running job {job.name}")
        else:
            click.echo("")

        job_suuid = job.short_uuid

    try:
        run = askanna_run.start(
            job_suuid=job_suuid,
            data=data,
            name=name,
            description=description,
        )
    except Exception as e:
        click.echo(e)
        sys.exit(1)
    else:
        click.echo(
            "Succesfully started a new run for job '{}' in AskAnna".format(
                run.job.get("name")
            )
        )



import logging
import sys

import click
from askanna.core.push import push

logging.basicConfig(format="%(levelname)s:%(message)s", level=logging.INFO)

HELP = """
Command to make an ZIP archive of the current working folder.\n
Afterwards we send this ZIP archive to AskAnna.
"""

SHORT_HELP = "Push code to AskAnna"


@click.command(help=HELP, short_help=SHORT_HELP)
@click.option("--force", "-f", is_flag=True, help="Force push")
@click.option(
    "--description",
    "-d",
    required=False,
    type=str,
    help="Add description to this code",
    default="",
)
@click.option(
    "--message",
    "-m",
    required=False,
    type=str,
    help="Add description to this code",
    default="",
)
def cli(force, description, message):
    if len(description) > 0 and len(message) > 0:
        click.echo("Cannot use both --description and --message.", err=True)
        sys.exit(1)
    elif len(message) > 0:
        description = message

    push(force, description)



# -*- coding: utf-8 -*-
import math
import os
import sys

import click

from askanna.core import client as askanna_client
from askanna.core.utils import diskunit


class ChunkedDownload:
    def __init__(self, url : str):
        """
        Takes an url to download from, this can be a url which could redirect to another URI.
        """
        self.history = []
        self.download_queue = []
        self.chunk_size = 10 * diskunit.MiB

        self.url = url
        self.size = 0
        self.accept_ranges = "none"

        self.client = askanna_client
        self.perform_preflight(url=url)

    @property
    def status_code(self):
        """
        Return the last status_code
        """
        return self.history[-1][1]

    def perform_preflight(self, url):
        """
        We take the session from self.client and will do a preflight request to
        determine whether the url will result in a (final) http_response_code=200
        """
        r = self.client.head(url)
        self.history.append([url, r.status_code, r.headers])

        if r.status_code in [301, 302] and r.headers.get("Location"):
            self.url = r.headers.get("Location")
            return self.perform_preflight(url=r.headers.get("Location"))
        elif r.status_code == 200:
            self.size = int(r.headers.get("Content-Length"))
            self.content_type = r.headers.get("Content-Type")
            self.accept_ranges = r.headers.get("Accept-Ranges", "none")

    def setup_download(self):
        """
        Download the self.url and chunk (if needed) the download
        """
        no_chunks = math.ceil(self.size / self.chunk_size)
        for chunk_no in range(0, no_chunks):
            self.download_queue.append(
                [
                    chunk_no,
                    chunk_no * self.chunk_size,
                    (chunk_no + 1) * self.chunk_size - 1,
                ]
            )
        # fix last end byte
        self.download_queue[-1][-1] = self.size

    def download(self, output_file : str):
        """
        Download the target_url, we consume the queue and write partial files.
        In the end we glue them to the original intended file.
        """
        self.setup_download()
        finished_queue = []
        while len(self.download_queue):
            chunk = self.download_queue.pop(0)
            range_header = {
                "Range": "{accept_ranges}={start}-{end}".format(
                    accept_ranges=self.accept_ranges, start=chunk[1], end=chunk[2]
                )
            }

            try:
                r = self.client.get(self.url, stream=True, headers=range_header)
                if not r.content or r.status_code != 206:
                    click.echo(f"HTTP response status code: {r.status_code}", err=True)
                    click.echo("Something went wrong with downloading the file content. Please try again.", err=True)
                    sys.exit(1)

                # so far so good on the download, save the result
                with open("file_{}.part".format(chunk[0]), "wb") as f:
                    for dlchunk in r.iter_content(chunk_size=1024):
                        f.write(dlchunk)
            except Exception as e:
                click.echo(e)
                self.download_queue.append(chunk)
            else:
                finished_queue.append(chunk)

        # combine the chunks in to one file
        with open(output_file, "wb") as f:
            finished_queue = sorted(finished_queue, key=lambda x: x[0])
            for chunk in finished_queue:
                chunkfilename = "file_{}.part".format(chunk[0])
                with open(chunkfilename, "rb") as chunkfile:
                    f.write(chunkfile.read())
                # delete the chunkfile
                os.remove(chunkfilename)



from .base import BaseCLItest


class TestCliPush(BaseCLItest):
    """
    askanna push

    We expect to initiate a push action of our code to the AskAnna server
    """
    verb = "push"

    def test_command_push_base(self):
        assert "push" in self.result.output
        self.assertIn("push", self.result.output)
        self.assertNotIn("noop", self.result.output)



"""
Core functions for management of runs in AskAnna
This is the class which act as gateway to the API of AskAnna
"""
import click

from typing import List

from askanna.core import client, exceptions
from askanna.core.dataclasses import Run, RunInfo, RunStatus
from askanna.core.job import JobGateway
from askanna.core.metrics import MetricGateway
from askanna.core.variables_tracked import VariableTrackedGateway


class RunGateway:
    def __init__(self, *args, **kwargs):
        self.client = client
        self.run_suuid = None

        # instantiated gateways for query metrics and variables
        self.metrics = MetricGateway()
        self.variables = VariableTrackedGateway()

    def start(
        self,
        job_suuid: str = None,
        data: dict = None,
        name: str = None,
        description: str = None,
    ) -> Run:

        url = "{}{}/{}/".format(self.client.config.remote, "run", job_suuid)

        r = self.client.post(
            url,
            json=data,
            params={
                "name": name,
                "description": description,
            },
        )
        if r.status_code != 200:
            raise exceptions.PostError(
                "{} - Something went wrong while starting a "
                "run: {}".format(r.status_code, r.reason)
            )
        run = Run(**r.json())
        self.run_suuid = run.short_uuid

        return run

    def list(self, query_params: dict = None) -> List[RunInfo]:
        url = "{}{}/".format(
            self.client.config.remote,
            "runinfo",
        )
        r = self.client.get(url, params=query_params)
        if r.status_code != 200:
            raise exceptions.GetError(
                "{} - Something went wrong while retrieving the status "
                "of the run: {}".format(r.status_code, r.reason)
            )
        return [RunInfo(**r) for r in r.json().get("results")]

    def detail(self, suuid: str = None) -> RunInfo:
        """
        Retrieve details on a Run
        The suuid is optional and can be retrieved from `self.run_suuid`
        if the gateway was instantiated with it.

        returns RunInfo
        """
        suuid = suuid or self.run_suuid

        url = "{}{}/{}/".format(self.client.config.remote, "runinfo", suuid)

        r = self.client.get(url)
        if r.status_code != 200:
            raise exceptions.GetError(
                "{} - Something went wrong while retrieving the status "
                "of the run: {}".format(r.status_code, r.reason)
            )

        return RunInfo(**r.json())

    def status(self, suuid: str = None) -> RunStatus:
        suuid = suuid or self.run_suuid

        if not suuid:
            raise exceptions.RunError(
                "There is no run SUUID provided. Did you start a run?"
            )

        url = "{}{}/{}/".format(self.client.config.remote, "status", suuid)

        r = self.client.get(url)
        if r.status_code != 200:
            raise exceptions.GetError(
                "{} - Something went wrong while retrieving the status "
                "of the run: {}".format(r.status_code, r.reason)
            )

        return RunStatus(**r.json())


class RunActionGateway:
    """
    This provides to query runs, using the RunGateway
    the .get() returns eiter
    """

    multiple = False
    run_suuid = None

    def __init__(self):
        self.gateway = RunGateway()

    def start(
        self,
        job_suuid: str = None,
        data: dict = None,
        job_name: str = None,
        project_suuid: str = None,
        name: str = None,
        description: str = None,
    ) -> Run:
        if not job_suuid and not job_name:
            raise exceptions.PostError(
                "To start a run we need at least a job SUUID or job name"
            )
        elif not job_suuid:
            if not project_suuid:
                project_suuid = self.gateway.client.config.project_suuid
            job_gateway = JobGateway()
            job_suuid = job_gateway.get_job_by_name(
                job_name=job_name, project_suuid=project_suuid
            ).short_uuid

        run = self.gateway.start(
            job_suuid=job_suuid,
            data=data,
            name=name,
            description=description,
        )
        self.run_suuid = run.short_uuid
        return run

    def status(self, suuid: str = None) -> Run:
        suuid = suuid or self.run_suuid
        return self.gateway.status(suuid=suuid)

    def get(
        self,
        run,
        include_metrics: bool = False,
        include_variables: bool = False,
    ) -> RunInfo:
        if not run:
            click.echo("Please specify the run SUUID to 'get' a run", err=True)
            return

        runinfo = self.gateway.detail(suuid=run)
        if include_metrics:
            # also fetch the metrics for the runs
            metrics = self.gateway.metrics.get(run=runinfo.short_uuid)
            # run.metrics
            runinfo.metrics = metrics

        if include_variables:
            # also fetch the variables for the runs
            variables = self.gateway.variables.get(run=runinfo.short_uuid)
            # add the metrics to the runobjects
            # run.variables
            runinfo.variables = variables
        return runinfo


class RunMultipleQueryGateway(RunActionGateway):
    def get_query(self, project: str = None, job: str = None, runs: list = None):
        """
        We return jobs/ metrics from either:
        - project
        - job
        - runs (specific run_suuids)

        When runs is set, this takes precedence over project and job.
        When runs is empty, look at project or job, where job takes precedence
        """
        query = {}
        if runs and len(runs):
            # build comma separated string from runs if more then 1 run
            query = {"runs": ",".join(runs)}
        if project:
            query.update({"project": project})
        if job:
            query.update({"job": job})
        return query

    def get(
        self,
        project: str = None,
        job: str = None,
        job_name: str = None,
        runs: list = None,
        limit: int = 100,
        offset: int = 0,
        ordering: str = "-created",
        include_metrics: bool = False,
        include_variables: bool = False,
    ) -> List[RunInfo]:
        if job and job_name:
            raise exceptions.GetError(
                "Parameters 'job' and 'job_name' are both set. Please use or 'job' or 'job_name'."
            )
        if job_name:
            project_suuid = project or self.gateway.client.config.project_suuid
            job_gateway = JobGateway()
            job = job_gateway.get_job_by_name(
                job_name=job_name, project_suuid=project_suuid
            ).short_uuid

        query = self.get_query(project, job, runs)
        query.update(
            {
                "limit": limit,
                "offset": offset,
                "ordering": ordering,
            }
        )
        rgw = RunGateway()
        runs = rgw.list(query_params=query)
        if include_metrics:
            # also fetch the metrics for the runs
            for run in runs:
                metrics = self.gateway.metrics.get(run=run.short_uuid)
                # run.metrics
                run.metrics = metrics

        if include_variables:
            # also fetch the variables for the runs
            for run in runs:
                variables = self.gateway.variables.get(run=run.short_uuid)
                # run.metrics
                run.variables = variables

        return runs



from dataclasses import dataclass
import json
import os
import sys
import tempfile
from typing import Any, List
import uuid

import click
from dateutil import parser as dateutil_parser

from askanna.core import client, exceptions
from askanna.core.dataclasses import Metric, MetricDataPair, MetricLabel
from askanna.core.utils import (
    create_suuid,
    json_serializer,
    labels_to_type,
    prepare_and_validate_value,
    translate_dtype,
)

__all__ = ["track_metric", "track_metrics", "MetricCollector", "MetricGateway"]

"""
How to use metrics module:

from askanna.core.metrics import *

mc = MetricCollector()
mc.add(metric=Metric(metric=MetricDataPair(name="foo", value=1, dtype="int"), label=[]))

# get json:

mc.to_json()
"""


class MetricsQuerySet:
    metrics = []
    queries = []
    parent = None

    def __init__(self, metrics: list = [], queries: list = [], parent=None):
        self.metrics = metrics
        self.queries = queries
        self.parent = parent

    def __len__(self):
        return len(self.metrics)

    def __iter__(self):
        yield from self.metrics

    def get(self, name, **kwargs):
        ret = self.filter(name, **kwargs)
        if len(ret) == 1:
            return ret[0]
        elif len(ret) > 1:
            click.echo(
                f"Found more metrics matching name '{name}',"
                f" use .filter(name='{name}') to get all matches.",
                err=True,
            )
            return ret[0]
        else:
            click.echo(f"A metric with name '{name}' is not found")

    def filter(self, name, **kwargs):
        ret = list(
            filter(lambda x: x.get("metric", {}).get("name") == name, self.metrics)
        )
        return ret

    def to_json(self) -> str:
        return json.dumps(self.metrics, indent=1)

    def to_raw(self) -> list:
        return self.metrics


@dataclass
class MetricCollector:
    metrics = []
    run_suuid: str = None
    local_suuid: str = None
    metrics_folder: str = None
    metrics_file: str = None
    changed: bool = False

    def __post_init__(self):
        if not self.run_suuid:
            # generate local suuid
            local_uuid = uuid.uuid4()
            self.local_suuid = create_suuid(local_uuid)

        self.metrics_folder = self.get_metrics_filepath()
        self.metrics_file = os.path.join(self.metrics_folder, "metrics.json")

        if self.run_suuid:
            self.restore_session()

    def get_suuid(self):
        """
        Return the run_suuid if exist, otherwise the local_suuid
        """
        return self.run_suuid or self.local_suuid

    def get_metrics_filepath(self):
        tmpdir = tempfile.gettempdir()
        return os.path.join(tmpdir, "askanna", "run", self.get_suuid())

    def metric_to_type(self, metric: dict = None) -> MetricDataPair:
        return MetricDataPair(
            name=metric["name"], value=metric["value"], dtype=metric["type"]
        )

    def label_to_type(self, label_list: list = None) -> list:
        _ret = []
        for label in label_list:
            _ret.append(
                MetricLabel(
                    name=label["name"], value=label["value"], dtype=label["type"]
                )
            )
        return _ret

    def restore_session(self):
        """
        We can only restore a session when the self.run_uuid is set
        """
        if not self.run_suuid:
            return

        restored_metrics = []
        try:
            with open(self.metrics_file, "r") as f:
                stored_metrics = json.load(f)
                for stored_metric in stored_metrics:
                    restored_metrics.append(
                        Metric(
                            metric=self.metric_to_type(stored_metric.get("metric")),
                            label=self.label_to_type(stored_metric.get("label")),
                            run_suuid=stored_metric.get("run_suuid"),
                            created=dateutil_parser.isoparse(
                                stored_metric.get("created")
                            ),
                        )
                    )

                self.metrics = restored_metrics
                del restored_metrics
        except FileNotFoundError:
            pass

    def save_session(self):
        # create folder for this session, name is the run_suuid with prefix askanna-metrics

        os.makedirs(self.metrics_folder, exist_ok=True)
        with open(self.metrics_file, "w") as f:
            f.write(self.to_json())

    def has_metrics(self) -> bool:
        return len(self.metrics) > 0

    def add(self, metric: Metric) -> None:
        metric.run_suuid = self.get_suuid()
        self.metrics.append(metric)
        self.save_session()
        self.changed = True

    def to_dict(self, yes=True) -> List:
        return [m.to_dict() for m in self.metrics]

    def to_json(self, yes=True) -> str:
        """
        We serialize all metrics to json for reporting or sending to AskAnna
        """
        return json.dumps(self.to_dict(), default=json_serializer)

    def save(self, run_suuid: str = None, force: bool = False) -> None:
        """
        Connect to AskAnna backend to save the metrics
        We will only attempt to submit if there is data recorded
        """
        if not self.changed and not force:
            # We will not save when there is nothing changed (e.g. no new metric added)
            return
        if len(self.to_dict()) and not run_suuid and not self.run_suuid:
            click.echo(
                "The run SUUID is not set for this session.\n"
                + "AskAnna cannot submit the metrics to the platform.",
                err=True,
            )
            click.echo(
                f"Your metrics are saved locally in:\n{self.metrics_file}", err=True
            )
            sys.exit(1)
        self.save_session()
        mgw = MetricGateway()
        # the run_suuid takes higher prio then the one set in init
        short_uuid = run_suuid or self.run_suuid
        mgw.change(short_uuid=short_uuid, metrics=self.to_dict())


# Start logic for metric collection
mc = MetricCollector(run_suuid=os.getenv("AA_RUN_SUUID"))


def track_metric(name: str, value: Any, label: dict = None) -> None:
    # store the metric
    if value:
        value, valid = prepare_and_validate_value(value)
        if not valid:
            click.echo(
                f"AskAnna cannot store this datatype. Metric not stored for {name}, {value}, {label}.",
                err=True
            )
            return
    # add value to track queue
    if value:
        datapair = MetricDataPair(name=name, value=value, dtype=translate_dtype(value))
    else:
        datapair = MetricDataPair(name=name, value=None, dtype="tag")
    labels = labels_to_type(label, labelclass=MetricLabel)

    metric = Metric(metric=datapair, label=labels)
    mc.add(metric)


def track_metrics(metrics: dict, label: dict = None) -> None:
    """
    Transform many metrics to internal format
    """
    for name, value in metrics.items():
        track_metric(name, value, label)


class MetricGateway:
    def __init__(self, *args, **kwargs):
        self.client = client

    def get(self, run: str = None, job: str = None) -> MetricsQuerySet:
        """
        Depending on what we get as argument, we call the 'list' function
        with specific query params for a different endpoint
        """

        if not any([run, job]):
            # we try to get local metrics if this exists

            if mc.run_suuid and mc.has_metrics():
                return MetricsQuerySet(metrics=mc.to_dict())
            elif mc.run_suuid:
                # RUN_SUUID is set but we don't have metrics
                return MetricsQuerySet(metrics=[])
            else:
                click.echo(
                    "We cannot find a run metric file for the active run.", err=True
                )
        else:
            pre_query_params = {
                "run_suuid": run,
                "job_suuid": job,
            }
            # filter out the none's:
            query_params = dict(
                filter(lambda x: x[1] is not None, pre_query_params.items())
            )
            return self.list(query_params=query_params)

    def list(self, query_params: dict = None) -> MetricsQuerySet:
        """
        List endpoint order:
        - job
        - run
        """

        endpoints = {
            "job": lambda x: "{}job/{}/metrics/".format(self.client.config.remote, x),
            "run": lambda x: "{}runinfo/{}/metrics/".format(
                self.client.config.remote, x
            ),
        }
        if "run_suuid" in query_params.keys():
            url = endpoints.get("run")(query_params.get("run_suuid"))
        elif "job_suuid" in query_params.keys():
            url = endpoints.get("job")(query_params.get("job_suuid"))

        r = self.client.get(url, params=query_params)
        return MetricsQuerySet(metrics=r.json())

    def change(self, short_uuid, metrics):
        url = "{}{}/{}/{}/{}/".format(
            self.client.config.remote, "runinfo", short_uuid, "metrics", short_uuid
        )

        r = self.client.put(url, json={"metrics": metrics})

        if r.status_code != 200:
            raise exceptions.GetError(
                "{} - Something went wrong while updating metrics "
                ": {}".format(r.status_code, r.json())
            )


# code for exiting in ipython from https://stackoverflow.com/a/40222538
# exit_register runs at the end of ipython %run or the end of the python interpreter
try:

    def exit_register(fun, *args, **kwargs):
        """Decorator that registers at post_execute. After its execution it
        unregisters itself for subsequent runs."""

        def callback():
            fun()
            ip.events.unregister("post_execute", callback)

        ip.events.register("post_execute", callback)

    ip = get_ipython()
except NameError:
    from atexit import register as exit_register


@exit_register
def handle_exit_save_metrics():
    try:
        if mc.has_metrics():
            mc.save()
    except Exception as e:
        # do a hard exit
        click.echo(f"Exit error in AskAnna SDK. {e}", err=True)
        os._exit(1)



from askanna.core.utils import extract_push_target, get_config
from askanna.core.user import User


config = get_config(check_config=False)


class Config:
    """
    Configuration management for AskAnna CLI & SDK
    """

    @property
    def remote(self):
        return config.get("askanna", {}).get(
            "remote", "https://beta-api.askanna.eu/v1/"
        )

    @property
    def user(self):
        token = config.get("auth", {}).get("token")
        return User(token=token)

    @property
    def push_target(self):
        return config.get("push-target")

    @property
    def workspace_suuid(self):
        try:
            push_target = extract_push_target(self.push_target)
        except ValueError:
            # the push-target is not set, so don't bother reading it
            return None
        return push_target.get("workspace_suuid")

    @property
    def project_suuid(self):
        try:
            push_target = extract_push_target(self.push_target)
        except ValueError:
            # the push-target is not set, so don't bother reading it
            return None
        return push_target.get("project_suuid")



# -*- coding: utf-8 -*-
import os
import sys

import click

from askanna.core import client as askanna_client
from askanna.core.utils import get_config


# read defaults from the environment
default_run_suuid = os.getenv("AA_RUN_SUUID")
default_payload_suuid = os.getenv("AA_PAYLOAD_SUUID")
default_payload_file = os.getenv("AA_PAYLOAD_PATH")

HELP = """Get the payload for the run which is stored in AskAnna"""
SHORT_HELP = "Get payload from AskAnna"


@click.command(help=HELP, short_help=SHORT_HELP)
@click.option("--run", "-r", default=default_run_suuid, show_default=True)
@click.option("--payload", "-p", default=default_payload_suuid, show_default=True)
@click.option(
    "--output", "-o", default=default_payload_file, show_default=True, type=click.Path()
)
def cli(run, payload, output):
    config = get_config()
    ASKANNA_API_SERVER = config["askanna"]["remote"]

    # we assume we can get jobrun id and the payload
    payload_url = (
        "{ASKANNA_API_SERVER}runinfo/{RUN_SUUID}/payload/{PAYLOAD_SUUID}/".format(
            RUN_SUUID=run,
            PAYLOAD_SUUID=payload,
            ASKANNA_API_SERVER=ASKANNA_API_SERVER,
        )
    )

    req = askanna_client.get(payload_url)

    if not req.status_code == 200:
        click.echo(f"Could not retrieve payload with SUUID: {payload}", err=True)
        sys.exit(1)

    if output:
        with open(output, "w") as f:
            f.write(req.text)
    else:
        click.echo(req.text)



import click
import os
import yaml

from askanna.core import client as askanna_client
from askanna.core.utils import get_config

HELP = """
This command will allow you to create an AskAnna project in your current directory
"""

SHORT_HELP = "Create a project in the current directory"


def get_user_info(server):

    url = "{server}rest-auth/user".format(server=server.replace("v1/", ""))
    ruser = askanna_client.get(url)
    if ruser.status_code == 200:
        return ruser.json()
    elif ruser.status_code == 401:
        print(
            "The provided token is not valid. Via `askanna logout` you can remove the token "
            "and via `askanna login` you can set a new token."
        )
    else:
        print("Could not connect to AskAnna at this moment")


class CreateProject:
    def __init__(self, name=None, user=None):
        self.config = get_config()
        self.api_server = self.config["askanna"]["remote"]
        self.name = name
        self.user = user or get_user_info(server=self.api_server)

    def find_workspace(self) -> list:
        """
        Find all workspaces where the user is part of and enumerate them
        """
        workspaces = []
        r = askanna_client.get("{}workspace/".format(self.api_server))

        for workspace in r.json():
            workspaces.append(
                {"suuid": workspace["short_uuid"], "name": workspace["name"]}
            )
        return workspaces

    def ask_workspace(self) -> str:
        """
        Ask which workspace the project will be created in.
        """
        workspaces = self.find_workspace()
        if len(workspaces) > 1:

            workspace_list_str = ""

            for idx, workspace in enumerate(workspaces, start=1):
                workspace_list_str += "%d. %s\n" % (idx, workspace["name"])

            workspace = click.prompt(
                "\nYou are a member of multiple workspaces. "
                + "In which workspace do you want to create the new project?\n"
                + workspace_list_str
                + "\n"
                + "Please enter the workspace number"
            )
            selected_workspace = workspaces[int(workspace) - 1]
        else:
            selected_workspace = workspaces[0]

        if click.confirm(
            "\nDo you want to create a project '{project}' in the workspace '{workspace}'?".format(
                project=self.name, workspace=selected_workspace["name"]
            ),
            abort=True,
        ):
            click.echo(
                "Thank you for the information. Anna will continue creating the project."
            )

        return selected_workspace["suuid"]

    def cli(self, workspace: str = None, description: str = None):
        if not self.name:
            name_user = ""
            if self.user.get("name"):
                name_user = " " + self.user.get("name")

            click.echo(
                f"Hi{name_user}! It is time to create a new project in AskAnna. "
                "We start with some information about the project."
            )
            click.echo("")
            self.name = click.prompt("Project name", type=str)

            if not description:
                description = click.prompt(
                    "Project description", type=str, default="", show_default=False
                )

        if not workspace:
            workspace = self.ask_workspace()

        r = askanna_client.post(
            self.api_server + "project/",
            data={
                "name": self.name,
                "workspace": workspace,
                "description": description,
            },
        )
        if r.status_code == 201:
            click.echo("\nYou have successfully created a new project in AskAnna!")
            return r.json()
        else:
            raise Exception("We could not create the project.")


@click.command(help=HELP, short_help=SHORT_HELP)
@click.argument("name", required=False)
@click.option(
    "--workspace",
    "-w",
    envvar="WORKSPACE_SUUID",
    show_default=True,
    help="Workspace SUUID where you want to create the project",
)
@click.option("--description", "-d", help="Description of the project")
def cli(name, workspace, description):
    cwd = os.getcwd()
    askanna_project_file = os.path.join(cwd, "askanna.yml")
    if os.path.exists(askanna_project_file):
        click.echo(
            "This directory already has an 'askanna.yml' file. If you continue, we will "
            "create a new project in AskAnna."
        )
        click.echo(
            "But, we will not update your 'askanna.yml' file with the push-target of the "
            "new project. If you continue, you need to add the push-target yourself.\n"
        )
        click.confirm(
            "Do you want to continue without updating the 'askanna.yml'?", abort=True
        )
        click.echo("")

    project_creator = CreateProject(name=name)
    project_info = project_creator.cli(workspace, description)

    if not os.path.exists(askanna_project_file):
        with open(askanna_project_file, "w") as pf:
            pf.write(yaml.dump({"push-target": project_info["url"]}, indent=2))
    else:
        click.echo(
            "\nTo be able to push your code to AskAnna, you need to update your local `askanna.yml` file. "
            "In the `askanna.yml` set or update the push-target with:"
        )
        click.echo("push-target: {}".format(project_info["url"]))

    # finish
    click.echo("\nCheck your project in AskAnna at:")
    click.echo("{project_url}".format(project_url=project_info["url"]))
    click.echo(
        "\nAs a first step you can configure your first job for this project in the `askanna.yml` file. "
        "Ones you are done, you can easily push your code to AskAnna via:\n"
        "\n"
        "  askanna push\n"
        "\n"
        "Success with your project!"
    )



import json

from askanna import (
    __version__ as askanna_version,
    USING_ASKANNA_CLI,
)
from .config import Config
from .session import Session
from .utils import json_serializer


class Client:
    """
    Client for communication with AskAnna API service
    """

    def __init__(self, headers=None, *args, **kwargs):
        self.config = Config()
        self.headers = headers or self.generate_authenication_header()
        self.session = Session(headers=self.headers)

    def generate_authenication_header(self):
        if not self.config.user.token:
            return {}

        askanna_agent = USING_ASKANNA_CLI and "cli" or "python-sdk"

        return {
            "askanna-agent": askanna_agent,
            "askanna-agent-version": askanna_version,
            "user-agent": f"askanna-python/{askanna_version}",
            "Authorization": f"Token {self.config.user.token}",
        }

    def get(self, url, **kwargs):
        return self.session.get(url, **kwargs)

    def head(self, url, **kwargs):
        return self.session.head(url, **kwargs)

    def patch(self, url, **kwargs):
        if kwargs.get("json"):
            kwargs["json"] = json.loads(
                json.dumps(kwargs.get("json"), default=json_serializer)
            )
        return self.session.patch(url, **kwargs)

    def put(self, url, **kwargs):
        if kwargs.get("json"):
            kwargs["json"] = json.loads(
                json.dumps(kwargs.get("json"), default=json_serializer)
            )
        return self.session.put(url, **kwargs)

    def post(self, url, **kwargs):
        if kwargs.get("json"):
            kwargs["json"] = json.loads(
                json.dumps(kwargs.get("json"), default=json_serializer)
            )
        return self.session.post(url, **kwargs)

    def create(self, url, **kwargs):
        if kwargs.get("json"):
            kwargs["json"] = json.loads(
                json.dumps(kwargs.get("json"), default=json_serializer)
            )
        return self.session.create(url, **kwargs)

    def delete(self, url, **kwargs):
        if kwargs.get("json"):
            kwargs["json"] = json.loads(
                json.dumps(kwargs.get("json"), default=json_serializer)
            )
        return self.session.delete(url, **kwargs)



import unittest

from click.testing import CliRunner

from askanna.cli.tool import cli_commands


class BaseCLItest(unittest.TestCase):
    verb = ""

    def setUp(self):
        self.result = CliRunner().invoke(cli_commands, ['--help', self.verb, ])

    def testCommandLineAccess(self):
        assert self.result.exit_code == 0
        assert len(self.result.output) > 0



from dataclasses import dataclass, field
import datetime
from typing import Any, List, Dict
import uuid


@dataclass
class Job:
    name: str
    description: str
    uuid: uuid.UUID
    short_uuid: str
    project: dict
    notifications: dict
    schedules: list
    created: datetime.datetime
    modified: datetime.datetime
    environment: str
    timezone: str


@dataclass
class Project:
    name: str
    description: str
    uuid: uuid.UUID
    short_uuid: str
    created_by: dict
    package: dict
    notifications: dict
    template: str
    created: datetime.datetime
    modified: datetime.datetime
    workspace: dict
    url: str = None

    def __str__(self):
        return f"{self.name} {self.short_uuid}"


@dataclass
class Run:
    """
    This dataclass holds information about the run which just started
    """

    message_type: str
    status: str
    uuid: uuid.UUID
    short_uuid: str
    name: str
    next_url: str
    job: dict
    workspace: dict
    project: dict
    created: datetime.datetime
    updated: datetime.datetime
    finished: datetime.datetime = None
    duration: int = 0


@dataclass
class RunStatus:
    """
    This dataclass holds information about the status of a run.
    """

    message_type: str
    status: str
    uuid: uuid.UUID
    short_uuid: str
    name: str
    next_url: str
    job: dict
    workspace: dict
    project: dict
    environment: dict
    created: datetime.datetime
    updated: datetime.datetime
    finished: datetime.datetime = None
    duration: int = 0


@dataclass
class RunInfo:
    name: str
    description: str
    status: str
    uuid: uuid.UUID
    short_uuid: str

    project: dict
    artifact: dict
    package: dict
    owner: dict
    trigger: dict
    payload: dict
    jobdef: dict
    environment: dict

    metricsmeta: dict
    variablesmeta: dict

    created: datetime.datetime
    modified: datetime.datetime
    duration: int

    result: dict = None
    started: datetime.datetime = None
    finished: datetime.datetime = None

    metrics = []
    variables = []


@dataclass
class User:
    short_uuid: str
    name: str
    email: str
    is_active: bool
    date_joined: datetime.datetime
    last_login: datetime.datetime


@dataclass
class Variable:
    name: str
    value: str
    short_uuid: str
    uuid: uuid.UUID
    is_masked: bool
    created: datetime.datetime
    modified: datetime.datetime
    project: dict


@dataclass
class Workspace:
    name: str
    description: str
    uuid: uuid.UUID
    short_uuid: str
    created: datetime.datetime
    modified: datetime.datetime

    def __str__(self):
        return f"{self.name} {self.short_uuid}"


@dataclass
class MetricDataPair:
    name: str
    value: Any
    dtype: str

    def to_dict(self, yes=True) -> Dict:
        return {"name": self.name, "value": self.value, "type": self.dtype}


@dataclass
class MetricLabel:
    name: str
    value: Any
    dtype: str

    def to_dict(self, yes=True) -> Dict:
        return {"name": self.name, "value": self.value, "type": self.dtype}


@dataclass
class Metric:
    metric: MetricDataPair
    label: List[MetricLabel] = field(default_factory=list)
    run_suuid: str = None
    created: datetime.datetime = None

    def __post_init__(self):
        if not self.created:
            # record the created time always in UTC time so that we don't have
            # to figure out what local timezone is, this is always correct
            # and we don't have to calculate back to utc
            self.created = datetime.datetime.now(datetime.timezone.utc)

    def to_dict(self, yes=True) -> Dict:
        return {
            "metric": self.metric.to_dict(),
            "label": [label.to_dict() for label in self.label],
            "run_suuid": self.run_suuid,
            "created": self.created,
        }


@dataclass
class VariableDataPair:
    name: str
    value: Any
    dtype: str

    def to_dict(self, yes=True) -> Dict:
        return {"name": self.name, "value": self.value, "type": self.dtype}


@dataclass
class VariableLabel:
    name: str
    value: Any
    dtype: str

    def to_dict(self, yes=True) -> Dict:
        return {"name": self.name, "value": self.value, "type": self.dtype}


@dataclass
class VariableTracked:
    variable: VariableDataPair
    label: List[VariableLabel] = field(default_factory=list)
    run_suuid: str = None
    created: datetime.datetime = None

    def __post_init__(self):
        if not self.created:
            # record the created time always in UTC time so that we don't have
            # to figure out what local timezone is, this is always correct
            # and we don't have to calculate back to utc
            self.created = datetime.datetime.now(datetime.timezone.utc)

    def to_dict(self, yes=True) -> Dict:
        return {
            "variable": self.variable.to_dict(),
            "label": [label.to_dict() for label in self.label],
            "run_suuid": self.run_suuid,
            "created": self.created,
        }



class User:
    def __init__(self, *args, **kwargs):
        self.user = None
        self.token = kwargs.get('token')

    @property
    def authenticated(self):
        return self.token is not None



"""
Management of projects in AskAnna
This is the class which act as gateway to the API of AskAnna
"""
import sys

import click

from askanna.core import client, exceptions
from askanna.core.dataclasses import Project


class ProjectGateway:
    def __init__(self, *args, **kwargs):
        self.client = client
        self.workspace = kwargs.get("workspace")
        self.base_url = self.client.config.remote + "project/"

    def list(self, workspace_suuid: str = None) -> list:
        if not workspace_suuid and self.workspace:
            workspace_suuid = self.workspace.short_suuid

        if workspace_suuid:
            # build url to select for project only
            url = "{}{}/{}/{}".format(
                self.client.config.remote, "workspace", workspace_suuid, "projects"
            )
        else:
            url = "{}{}/".format(self.client.config.remote, "project")

        r = self.client.get(url)
        if r.status_code != 200:
            raise exceptions.GetError(
                "{} - Something went wrong while retrieving "
                "projects: {}".format(r.status_code, r.json())
            )

        return [Project(**project) for project in r.json()]

    def detail(self, suuid: str) -> Project:
        url = "{}{}/".format(self.base_url, suuid)
        r = self.client.get(url)

        if r.status_code != 200:
            raise exceptions.GetError(
                "{} - Something went wrong while retrieving "
                "project info: {}".format(r.status_code, r.json())
            )

        return Project(**r.json())

    def change(self, suuid: str, name: str = None, description: str = None) -> Project:
        """
        Change the project information
        """
        url = "{}{}/{}/".format(self.client.config.remote, "project", suuid)

        changes = [
            ["name", name],
            ["description", description],
        ]
        # filterout None's
        changes = list(filter(lambda x: x[1] is not None, changes))
        if len(changes) == 0:
            click.echo(
                "Nothing to change for this project. You did not provide a name or description.",
                err=True,
            )
            sys.exit(1)

        r = self.client.patch(url, json=dict(changes))
        if r.status_code == 200:
            project = Project(**r.json())
            click.echo(f"You have successfully changed the project: {project.name}")
            return project
        else:
            raise exceptions.PatchError(
                "{} - Something went wrong while updating the project information: "
                "{}".format(r.status_code, r.json())
            )



# -*- coding: utf-8 -*-
import click
import sys

from askanna import project as aa_project
from askanna.cli.utils import ask_which_project, ask_which_workspace
from askanna.core.config import Config
from askanna.core.utils import extract_push_target


config = Config()


@click.group()
def cli1():
    pass


@cli1.command(help="List projects available in AskAnna", short_help="List projects")
@click.option('--workspace', '-w', 'workspace_suuid', required=False, type=str,
              help='Workspace SUUID to list projects for a workspace')
def list(workspace_suuid):
    projects = aa_project.list(workspace_suuid)

    if not projects:
        click.echo("Based on the information provided, we cannot find any projects.")
        sys.exit(0)
    if workspace_suuid:
        click.echo(f"The projects for workspace \"{projects[0].workspace['name']}\" are:\n")
        click.echo("PROJECT SUUID          PROJECT NAME")
        click.echo("-------------------    -------------------------")
    else:
        click.echo("WORKSPACE SUUID        WORKSPACE NAME          PROJECT SUUID          PROJECT NAME")
        click.echo("-------------------    --------------------    -------------------    -------------------------")

    for project in sorted(projects, key=lambda x: (x.workspace['name'], x.name,)):
        if workspace_suuid:
            click.echo(
                "{project_suuid}    {project_name}".format(
                    project_suuid=project.short_uuid,
                    project_name=project.name[:25],
                )
            )
        else:
            click.echo(
                "{workspace_suuid}    {workspace_name}    {project_suuid}    {project_name}".format(
                    workspace_suuid=project.workspace['short_uuid'],
                    workspace_name="{:20}".format(project.workspace['name'])[:20],
                    project_suuid=project.short_uuid,
                    project_name=project.name[:25]
                )
            )


@click.group()
def cli2():
    pass


@cli2.command(help="Change project information in AskAnna", short_help="Change project")
@click.option("--id", "-i", "suuid", required=False, type=str, help="Project SUUID")
@click.option("--name", "-n", required=False, type=str, help="New name to set")
@click.option(
    "--description", "-d", required=False, type=str, help="New description to set"
)
def change(suuid, name, description):
    if not suuid:
        try:
            push_target = extract_push_target(config.push_target)
        except ValueError:
            # the push-target is not set, so don't bother reading it
            project_suuid = None
        else:
            project_suuid = push_target.get("project_suuid")

        if not project_suuid:
            workspace = ask_which_workspace(question="From which workspace do you want to change a project?")
            project = ask_which_project(question="Which project do you want to change?",
                                        workspace_suuid=workspace.short_uuid)
            project_suuid = project.short_uuid
        suuid = project_suuid

    if not name and not description:
        if click.confirm("\nDo you want to change the name of the project?"):
            name = click.prompt("New name of the project", type=str)
        if click.confirm("\nDo you want to change the description of the project?"):
            description = click.prompt("New description of the project", type=str)

        click.confirm("\nDo you want to change the project?", abort=True)

    aa_project.change(suuid=suuid, name=name, description=description)


cli = click.CommandCollection(
    sources=[
        cli1,
        cli2,
    ],
    help="Manage your projects in AskAnna",
    short_help="Manage projects in AskAnna",
)



# -*- coding: utf-8 -*-
import os
import re


def string_expand_variables(strings: list) -> list:
    var_matcher = re.compile(r"\$\{(?P<MYVAR>[\w\-]+)\}")
    for idx, line in enumerate(strings):
        matches = var_matcher.findall(line)
        for m in matches:
            line = line.replace("${" + m + "}", os.getenv(m.strip()))
        strings[idx] = line
    return strings



from .base import BaseCLItest


class TestCliLogout(BaseCLItest):
    """
    askanna logout

    We expect to logout the user (when authenicated)
    """
    verb = "logout"

    def testCommandLogoutBase(self):
        assert "logout" in self.result.output
        self.assertIn("logout", self.result.output)
        self.assertNotIn("noop", self.result.output)



# This file contains anything that needs to be loaded for `import askanna`. We separated this file to avoid conflicts
# while in installation mode

# Please note: never do `import askanna` here, this will cause an recursive import loop

import click

from askanna import USING_ASKANNA_CLI
from askanna.core.config import Config
from askanna.core.job import JobGateway
from askanna.core.metrics import track_metric, track_metrics, MetricGateway  # noqa
from askanna.core.variables_tracked import track_variable, track_variables  # noqa
from askanna.core.project import ProjectGateway
from askanna.core.result import ResultGateway
from askanna.core.run import RunMultipleQueryGateway, RunActionGateway
from askanna.core.utils import update_available
from askanna.core.workspace import WorkspaceGateway


if USING_ASKANNA_CLI:
    try:
        update_available()
    except Exception as e:
        click.echo(f"Something went wrong while checking if an update is available: {e}", err=True)

config = Config()

# Instantiated objects for query or actions, these do not contain any data on load and will be filled with data on
# using detail/get/list actions from it.
job = JobGateway()
metrics = MetricGateway()
project = ProjectGateway()
result = ResultGateway()
run = RunActionGateway()
runs = RunMultipleQueryGateway()
workspace = WorkspaceGateway()



import collections
import datetime
import glob
import ipaddress
import mimetypes
import os
from pathlib import Path
import re
import sys
from typing import Any, List, Tuple, Dict
import uuid
from zipfile import ZipFile

import click
import croniter
from email_validator import validate_email, EmailNotValidError
import igittigitt
import pytz
import tzlocal
import requests
import yaml
from yaml import load, dump

from askanna import __version__ as askanna_version
from askanna.settings import (
    CONFIG_FILE_ASKANNA,
    CONFIG_ASKANNA_REMOTE,
    DEFAULT_PROJECT_TEMPLATE,
    PYPI_PROJECT_URL,
)

try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper


from askanna.core import exceptions


StorageUnit = collections.namedtuple(
    "StorageUnit", ["B", "KiB", "MiB", "GiB", "TiB", "PiB"]
)

diskunit = StorageUnit(
    B=1, KiB=1024 ** 1, MiB=1024 ** 2, GiB=1024 ** 3, TiB=1024 ** 4, PiB=1024 ** 5
)

supported_data_types = {
    # primitive types
    "bool": "boolean",
    "int": "integer",
    "float": "float",
    "str": "string",
    # complex types
    "datetime.datetime": "datetime",
    "datetime.time": "time",
    "datetime.date": "date",
    "dict": "dictionary",
    "list": "list",
    "tag": "tag",
}

try:
    import numpy as np  # noqa: F401
except ImportError:
    # do nothing we don't support numpy
    numpy_available = False
else:
    numpy_available = True
    # list taken from https://numpy.org/doc/stable/user/basics.types.html
    numpy_types = {
        "numpy.bool_": "boolean",
        "numpy.intc": "integer",
        "numpy.int_": "integer",
        "numpy.uint": "integer",
        "numpy.short": "integer",
        "numpy.ushort": "integer",
        "numpy.longlong": "integer",
        "numpy.ulonglong": "integer",
        "numpy.half": "float",
        "numpy.float16": "float",
        "numpy.single": "float",
        "numpy.double": "float",
        "numpy.longdouble": "float",
        "numpy.csingle": "float",
        "numpy.cdouble": "float",
        "numpy.clongdouble": "float",
        # more platform specific types
        "numpy.int8": "integer",
        "numpy.int16": "integer",
        "numpy.int32": "integer",
        "numpy.int64": "integer",
        "numpy.uint8": "integer",
        "numpy.uint16": "integer",
        "numpy.uint32": "integer",
        "numpy.uint64": "integer",
        "numpy.intp": "integer",
        "numpy.uintp": "integer",
        "numpy.float32": "float",
        "numpy.float64": "float",
        "numpy.float_": "float",
        # python equivalant of Decimal, convert to float now
        "numpy.complex64": "float",
        "numpy.complex128": "float",
        "numpy.complex_": "float",
        # list type
        "numpy.array": "list",
    }
    supported_data_types.update(**numpy_types)


def object_fullname(o):
    # https://stackoverflow.com/a/2020083
    # o.__module__ + "." + o.__class__.__qualname__ is an example in
    # this context of H.L. Mencken's "neat, plausible, and wrong."
    # Python makes no guarantees as to whether the __module__ special
    # attribute is defined, so we take a more circumspect approach.
    # Alas, the module name is explicitly excluded from __qualname__
    # in Python 3.

    module = o.__class__.__module__
    if module is None or module == str.__class__.__module__:
        return o.__class__.__name__  # Avoid reporting __builtin__
    else:
        return module + "." + o.__class__.__name__


def flatten(t):
    return [item for sublist in t for item in sublist]


def update_available() -> bool:
    """
    Check whether most recent release of AskAnna on PyPI is newer than the AskAnna version in use. If a newer version
    is available, return a info message with update instructions.
    """
    try:
        r = requests.get(PYPI_PROJECT_URL)
    except requests.exceptions.ConnectionError:
        return False
    except requests.exceptions.HTTPError:
        return False
    else:
        if r.status_code == 200:
            pypi_info = r.json()
        else:
            return False

    if askanna_version == pypi_info["info"]["version"]:
        return False
    else:
        click.echo(
            "[INFO] A newer version of AskAnna is available. Update via: pip install -U askanna"
        )
        return True


def check_for_project():
    """
    Performs a check if we are operating within a project folder. When
    we wish to perform a deploy action, we want to be on the same
    level with the `askanna.yml` to be able to package the file.
    """
    pyfiles = glob.glob("*.yml")

    # look for the setup.py file
    if "askanna.yml" in pyfiles:
        return True
    else:
        return False


def scan_config_in_path(cwd=None):
    """
    Look for askanna.yml in parent directories
    """
    if not cwd:
        cwd = os.getcwd()
    project_configfile = ""
    # first check whether we already can find in the current workdir
    if contains_configfile(cwd):
        project_configfile = os.path.join(cwd, "askanna.yml")
    else:
        # traverse up all directories untill we find an askanna.yml file
        split_path = os.path.split(cwd)
        # in any other cases, look in parent directories
        while split_path[1] != "":
            if contains_configfile(split_path[0]):
                project_configfile = os.path.join(split_path[0], "askanna.yml")
                break
            split_path = os.path.split(split_path[0])
    return project_configfile


def read_config(path: str) -> dict:
    """
    Reading existing config or return default dict
    """
    try:
        with open(os.path.expanduser(path), "r") as f:
            return load(f, Loader=Loader) or {}
    except FileNotFoundError:
        config_folder = os.path.dirname(CONFIG_FILE_ASKANNA)
        if not os.path.exists(config_folder):
            os.makedirs(config_folder, exist_ok=True)

        Path(CONFIG_FILE_ASKANNA).touch()

        # Write initial config if the config file didn't exist
        config = store_config(CONFIG_ASKANNA_REMOTE)
        with open(CONFIG_FILE_ASKANNA, "w") as f:
            f.write(config)

        return CONFIG_ASKANNA_REMOTE
    except TypeError as e:
        click.echo(e, err=True)
        sys.exit(1)
    except yaml.scanner.ScannerError as e:
        click.echo("Error reading askanna.yml due to:", err=True)
        click.echo(e.problem, err=True)
        click.echo(e.problem_mark, err=True)
        sys.exit(1)


def contains_configfile(path: str, filename: str = "askanna.yml") -> bool:
    return os.path.isfile(os.path.join(path, filename))


def get_config(check_config=True) -> Dict:
    config = read_config(CONFIG_FILE_ASKANNA)

    # overwrite the AA remote if AA_REMOTE is set in the environment
    is_remote_set = os.getenv("AA_REMOTE")
    if is_remote_set:
        config["askanna"] = config.get("askanna", {})
        config["askanna"]["remote"] = is_remote_set

    # if askanna remote is not set, add a default remote to the config file
    try:
        config["askanna"]["remote"]
    except KeyError:
        config = store_config(CONFIG_ASKANNA_REMOTE)
        with open(CONFIG_FILE_ASKANNA, "w") as f:
            f.write(config)
        config = read_config(CONFIG_FILE_ASKANNA)

    # overwrite the user token if AA_TOKEN is set in the environment
    is_token_set = os.getenv("AA_TOKEN")
    if is_token_set:
        config["auth"] = config.get("auth", {})
        config["auth"]["token"] = is_token_set

    # set the project template
    config["project"] = config.get("project", {})
    config["project"]["template"] = os.getenv(
        "PROJECT_TEMPLATE_URL", DEFAULT_PROJECT_TEMPLATE
    )

    project_config = scan_config_in_path()
    if project_config:
        config.update(**read_config(project_config))

    if check_config:
        validate_config(config)

    return config


def validate_config(config: Dict):
    try:
        config["auth"]["token"]
    except KeyError:
        click.echo(
            "You are not logged in. Please login first via `askanna login`.", err=True
        )
        sys.exit(1)


def store_config(new_config):
    original_config = read_config(CONFIG_FILE_ASKANNA)
    original_config.update(**new_config)
    output = dump(original_config, Dumper=Dumper)
    return output


# Zip the files that matches the filter from given directory
def zip_files_in_dir(directory_path: str, zip_file: ZipFile, ignore_file: str = None) -> None:
    files = get_files_in_dir(directory_path=directory_path, ignore_file=ignore_file)

    # Iterate over all the files and zip them
    for file in sorted(files):
        zip_file.write(file)


def get_files_in_dir(directory_path: str, ignore_file: str = None) -> set:
    file_list = set()

    ignore_parser = igittigitt.IgnoreParser()
    if ignore_file:
        ignore_parser.parse_rule_files(os.path.dirname(ignore_file), ignore_file)

    # Iterate over all the files in directory
    for root, _, files in os.walk(directory_path):
        for file in files:
            # Create complete filepath of file in directory
            file_path = os.path.join(root, file)
            if file_path.startswith("./"):
                file_path = file_path[2:]

            # can we add this file to the collection?
            if not ignore_parser.match(file_path):
                file_list.add(file_path)

    return file_list


def zip_paths(paths: List, zip_file: ZipFile, exclude_paths: List = []) -> None:
    files = get_files_in_paths(paths, exclude_paths)

    # Iterate over all the files and zip them
    for file in sorted(files):
        zip_file.write(file)


def get_files_in_paths(paths: List, exclude_paths: List = []) -> set:
    files = set()

    # first filter out empty string paths
    for path in filter(lambda x: len(x), paths):
        # replace asteriks if possible
        if path.startswith("*"):
            path = "." + path[1:]

        if path.endswith("/*"):
            path = path[:-1]

        if not os.path.exists(path):
            click.echo(f"{path} does not exists...skipping")
            continue

        # base folder path and check whether this is in the exclude_list
        base_folder = "/".join(path.split("/")[:2])
        if base_folder in exclude_paths:
            click.echo(
                f"[CONFIG ERROR] '{base_folder}' from '{path}' is on the exclusion list. AskAnna does not allow to "
                "access these files. This path is ignored and we continue with the other paths.",
                err=True,
            )
            continue

        if os.path.isdir(path):
            files.update(get_files_in_dir(path))
        else:  # so, we got a file
            if path.startswith("./"):
                path = path[2:]
            files.add(path)

    return sorted(files)


def create_zip_from_paths(filename: str, paths: List = []) -> None:
    """
    Create a ZipFile on the given `filename` location
    """
    # we exclude the following directories from included into the zip
    exclude_paths = [
        "/",
        "/bin",
        "/dev",
        "/lib",
        "/mnt",
        "/opt",
        "/proc",
        "/usr",
        "/var",
    ]

    with ZipFile(filename, mode="w") as f:
        zip_paths(paths, f, exclude_paths=exclude_paths)


def _file_type(path):
    """Mimic the type parameter of a JS File object.
    Resumable.js uses the File object's type attribute to guess mime type,
    which is guessed from file extention accoring to
    https://developer.mozilla.org/en-US/docs/Web/API/File/type.
    Parameters
    ----------
    path : str
        The path to guess the mime type of
    Returns
    -------
    str
        The inferred mime type, or '' if none could be inferred
    """
    type_, _ = mimetypes.guess_type(path)
    # When no type can be inferred, File.type returns an empty string
    return "" if type_ is None else type_


def getProjectInfo(project_suuid):
    # import the lib functions here, to prevent circular import
    from askanna.core import client
    from askanna.core.config import Config
    from askanna.core.dataclasses import Project

    config = Config()
    r = client.get(
        "{api_server}project/{project_suuid}/".format(
            api_server=config.remote, project_suuid=project_suuid
        ),
    )

    if r.status_code != 200:
        raise exceptions.GetError(
            "{} - Something went wrong while retrieving the "
            "project info: {}".format(r.status_code, r.reason)
        )

    return Project(**r.json())


def getProjectPackages(project, offset=0, limit=1):
    from askanna.core import client
    from askanna.core.config import Config

    config = Config()
    r = client.get(
        "{api_server}project/{project_suuid}/packages/?offset={offset}&limit={limit}".format(
            api_server=config.remote,
            project_suuid=project.short_uuid,
            offset=offset,
            limit=limit,
        )
    )
    if not r.status_code == 200:
        return []

    return r.json()


def extract_push_target(push_target: str):
    """
    Extract push target from the url configured
    Workspace is optional
    """
    if not push_target:
        raise ValueError("Cannot extract push-target if push-target is not set.")
    match_pattern = re.compile(
        r"(?P<http_scheme>https|http):\/\/(?P<askanna_host>[\w\.\-\:]+)\/(?P<workspace_suuid>[\w-]+){0,1}\/{0,1}project\/(?P<project_suuid>[\w-]+)\/{0,1}"  # noqa: E501
    )
    matches = match_pattern.match(push_target)
    matches_dict = matches.groupdict()
    return matches_dict


def validate_cron_line(cron_line: str) -> bool:
    """
    We validate the cron expression with croniter
    """
    return croniter.croniter.is_valid(cron_line)


def parse_cron_line(cron_line: str) -> str:
    """
    parse incoming cron definition
    if it is valid, then return the cron_line, otherwise a None
    """

    if isinstance(cron_line, str):
        # we deal with cron strings
        # first check whether we need to make a "translation" for @strings

        alias_mapping = {
            "@midnight": "0 0 * * *",
            "@yearly": "0 0 1 1 *",
            "@annually": "0 0 1 1 *",
            "@monthly": "0 0 1 * *",
            "@weekly": "0 0 * * 0",
            "@daily": "0 0 * * *",
            "@hourly": "0 * * * *",
        }
        cron_line = alias_mapping.get(cron_line.strip(), cron_line.strip())

    elif isinstance(cron_line, dict):
        # we deal with dictionary
        # first check whether we have valid keys, if one invalid key is found, return None
        valid_keys = set(["minute", "hour", "day", "month", "weekday"])
        invalid_keys = set(cron_line.keys()) - valid_keys
        if len(invalid_keys):
            return None
        cron_line = "{minute} {hour} {day} {month} {weekday}".format(
            minute=cron_line.get("minute", "0"),
            hour=cron_line.get("hour", "0"),
            day=cron_line.get("day", "*"),
            month=cron_line.get("month", "*"),
            weekday=cron_line.get("weekday", "*"),
        )

    if not validate_cron_line(cron_line):
        return None

    return cron_line


def parse_cron_schedule(schedule: List):
    """
    Determine and validate which format the cron defition it has, it can be one of the following:
    - * * * * *  (m, h, d, m, weekday)
    - @annotation (@yearly, @annually, @monthly, @weekly, @daily, @hourly)
    - dict (k,v with: minute,hour,day,month,weekday)
    """

    for cron_line in schedule:
        yield cron_line, parse_cron_line(cron_line)


def validate_yml_job_names(config):
    """
    Within AskAnna, we have several variables reserved and cannot be used for jobnames
    """
    reserved_keys = (
        "askanna",
        "cluster",
        "environment",
        "image",
        "job",
        "notifications",
        "project",
        "push-target",
        "timezone",
        "variables",
        "worker",
    )

    overlapping_with_reserved_keys = list(
        set(config.keys()).intersection(set(reserved_keys))
    )
    for overlap_key in overlapping_with_reserved_keys:
        is_dict = isinstance(config.get(overlap_key), dict)
        is_job = is_dict and config.get(overlap_key).get("job")
        if is_dict and is_job:
            # we do find a job definition under the name, so probably a jobdef
            click.echo(
                f"The name `{overlap_key}` cannot be used for a job.\n"
                "This name is used to configure something else in AskAnna.\n"
                f"Invalid job name: {overlap_key}",
                err=True,
            )
            return False
    return True


def validate_yml_environments(config: Dict, jobname=None) -> bool:
    """
    Validate the environment definitions (if defined)
    """
    # Do we have a global `environment` defined?
    environment = config.get("environment")
    if environment:
        if not isinstance(environment, dict):
            if jobname:
                click.echo(
                    f"Invalid definition of `environment` found in job `{jobname}`:\n"
                    f"environment: {environment}\n"
                    "\n"
                    "For environment documentation: https://docs.askanna.io/environments/",
                    err=True,
                )
            else:
                click.echo(
                    "Invalid definition of `environment` found:\n"
                    f"environment: {environment}\n"
                    "\n"
                    "For environment documentation: https://docs.askanna.io/environments/",
                    err=True,
                )
            return False
        else:
            # make sure we have at least the `image` defined
            image = environment.get("image")
            if not image:
                click.echo(
                    "`image` was not defined in the `environment`:\n"
                    f"environment: {environment}\n"
                    "\n"
                    "For environment documentation: https://docs.askanna.io/environments/",
                    err=True,
                )
                return False

    return True


def validate_yml_notifications(config: Dict, jobname=None) -> bool:
    """
    Validate the definition of `notifications` in global and job
    """
    notifications = config.get("notifications")
    if notifications and not isinstance(notifications, dict):
        if jobname:
            click.echo(
                f"Invalid definition of `notifications` found in job `{jobname}`:\n"
                f"notifications: {notifications}"
                "\n"
                "For notifications documentation: https://docs.askanna.io/jobs/notifications/",
                err=True,
            )
        else:
            click.echo(
                "Invalid `notifications` setting found:"
                f"notifications: {notifications}"
                "\n"
                "For notifications documentation: https://docs.askanna.io/jobs/notifications/",
                err=True,
            )
        return False
    elif notifications and isinstance(notifications, dict):
        # check for values of `all` and `error`, both should be dicts
        noti_all = notifications.get("all")
        noti_error = notifications.get("error")
        if any(
            [
                noti_all and not isinstance(noti_all, dict),
                noti_error and not isinstance(noti_error, dict),
            ]
        ):
            if jobname:
                click.echo(
                    f"Invalid definition of `notifications` found in job `{jobname}`:\n"
                    f"notifications: {notifications}"
                    "\n"
                    "For notifications documentation: https://docs.askanna.io/jobs/notifications/",
                    err=True,
                )
            else:
                click.echo(
                    "Invalid `notifications` setting found:\n"
                    f"notifications: {notifications}"
                    "\n"
                    "For notifications documentation: https://docs.askanna.io/jobs/notifications/",
                    err=True,
                )
            return False

        # proceed with validation of the content
        # assume we can get values out of all/email and error/email
        noti_all_values = notifications.get("all", {}).get("email", [])
        noti_error_values = notifications.get("error", {}).get("email", [])

        all_email = sorted(
            flatten(
                list(
                    map(
                        lambda x: x.split(","),
                        noti_all_values + noti_error_values,
                    )
                )
            )
        )
        for value in all_email:
            if value.startswith("${") and value.endswith("}"):
                # we found a variable not subsituted yet
                continue
            if value in ["workspace admins", "workspace members"]:
                # valid addresees
                continue

            # validate on actual content of the value whether it is an e-mail address or not
            try:
                validate_email(value)
            except EmailNotValidError:
                click.echo(
                    "Invalid `notifications/email` found:\n" f"{value}" "\n",
                    err=True,
                )
                return False

    return True


def validate_askanna_yml(config):
    """
    Given an dictionary of askanna.yml validate each component
    """
    jobs = config.items()

    global_timezone = config.get("timezone")
    # validate the global timezone
    if global_timezone:
        if global_timezone not in pytz.all_timezones:
            click.echo(
                "Invalid timezone setting found in askanna.yml:\n"
                + f"timezone: {global_timezone}",
                err=True,
            )
            return False
        timezone_checked = True  # used later, so we only print a warning message about the timezone once
    else:
        timezone_checked = False  # used later, so we only print a warning message about the timezone once

    # validate whether the global environment definitions are correct
    if not validate_yml_environments(config):
        return False

    # validate for global notification settings
    if not validate_yml_notifications(config):
        return False

    # validate jobs names
    if not validate_yml_job_names(config):
        return False

    for jobname, job in jobs:
        if isinstance(job, dict):
            if not validate_yml_environments(job, jobname=jobname):
                return False
            if not validate_yml_notifications(job, jobname=jobname):
                return False

            schedule = job.get("schedule")
            timezone = job.get("timezone")
            if not schedule:
                continue
            # validate the schedule
            for cron_line, parsed in parse_cron_schedule(schedule):
                if not parsed:
                    click.echo(
                        f"Invalid schedule definition found in job `{job}`: {cron_line}",
                        err=True,
                    )
                    return False
            # validate the timezone if set
            timezone = job.get("timezone")
            if timezone and timezone not in pytz.all_timezones:
                click.echo(
                    f"Invalid timezone setting found in job `{jobname}`:\n"
                    + f"timezone: {timezone}",
                    err=True,
                )
                return False

            # validate the schedule
            schedule = job.get("schedule")
            if schedule:
                for cron_line, parsed in parse_cron_schedule(schedule):
                    if not parsed:
                        click.echo(
                            f"Invalid schedule definition `{cron_line}` found in job `{jobname}`",
                            err=True,
                        )
                        return False

                if not timezone and not global_timezone and not timezone_checked:
                    timezone_local = getLocalTimezone()
                    if timezone_local != "UTC":
                        click.echo(  # noqa
                            f"""
By default, the AskAnna platform uses time zone UTC. Your current time zone is {timezone_local}.
To use your local time zone for scheduling jobs in this project, add the next line to your config in `askanna.yml`:

timezone: {timezone_local}

For more information, read the documentation: https://docs.askanna.io/jobs/create-job/#time-zone
"""
                        )
                    timezone_checked = True
    return True


# generation of suuid
def bx_encode(n, alphabet):
    """
    Encodes an integer :attr:`n` in base ``len(alphabet)`` with
    digits in :attr:`alphabet`.

    ::
        # 'ba'
        bx_encode(3, 'abc')
    :param n:            a positive integer.
    :param alphabet:     a 0-based iterable.
    """

    if not isinstance(n, int):
        raise TypeError("an integer is required")

    base = len(alphabet)

    if n == 0:
        return alphabet[0]

    digits = []

    while n > 0:
        digits.append(alphabet[n % base])
        n = n // base

    digits.reverse()
    return "".join(digits)


def str_to_suuid(string, n) -> str:
    return [string[i : i + n] for i in range(0, len(string), n)]


def create_suuid(uuid_obj) -> str:
    """
    Given an uuid4, return the suuid form of it
    """
    alphabet = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"

    token_length = 16
    group_size = 4
    groups = int(token_length / group_size)

    # Generate a random UUID if not given
    if not uuid_obj:
        uuid_obj = uuid.uuid4()

    # Convert it to a number with the given alphabet,
    # padding with the 0-symbol as needed)
    token = bx_encode(int(uuid_obj.hex, 16), alphabet)
    token = token.rjust(token_length, alphabet[0])

    return "-".join(str_to_suuid(token, group_size)[:groups])


def serialize_numpy_for_json(obj):
    # the o.item() is a generic method on each numpy dtype.
    if isinstance(obj, np.generic):
        return obj.item()
    if isinstance(obj, np.ndarray):
        return obj.tolist()
    return obj


def json_serializer(obj):
    if isinstance(obj, (datetime.time, datetime.date, datetime.datetime)):
        return obj.isoformat()
    if numpy_available:
        obj = serialize_numpy_for_json(obj)
    return obj


def translate_dtype(value: Any) -> str:
    """
    Return the full name of the type, if not listed, return the typename from the input
    """
    typename = object_fullname(value)

    return supported_data_types.get(typename, typename)


def validate_value(value: Any) -> bool:
    """
    Validate whether the value set is supported
    """

    return object_fullname(value) in supported_data_types


def transform_value(value: Any) -> Tuple[Any, bool]:
    """
    Transform values in support datatypes
    """
    if object_fullname(value) == "range":
        return list(value), True

    return value, False


def prepare_and_validate_value(value: Any) -> Tuple[Any, bool]:
    """
    Validate value and if necessary transform values in support datatypes
    """
    if validate_value(value):
        return value, True

    # Try to transform the value
    value, transform = transform_value(value)
    if transform:
        return value, True

    return value, False


def labels_to_type(label: Any = None, labelclass=collections.namedtuple) -> List:
    # process labels
    labels = []

    # test label type, if it is a string, then convert it to tag
    if isinstance(label, str):
        labels.append(labelclass(name=label, value=None, dtype="tag"))
    elif isinstance(label, list):
        for name in label:
            labels.append(labelclass(name=name, value=None, dtype="tag"))
    elif label:
        for k, v in label.items():
            if v is None:
                labels.append(labelclass(name=k, value=None, dtype="tag"))
            else:
                if v:
                    v, valid = prepare_and_validate_value(v)
                    if valid:
                        labels.append(
                            labelclass(name=k, value=v, dtype=translate_dtype(v))
                        )
                    else:
                        click.echo(
                            f"AskAnna cannot store this datatype. Label {k} with value {v} not stored.",
                            err=True,
                        )
    return labels


def isIPAddress(ip: str) -> bool:
    try:
        ipaddress.ip_address(ip)
    except ValueError:
        return False
    return True


def getLocalTimezone() -> str:
    """
    Determine the local timezone name
    """
    return tzlocal.get_localzone().zone


def content_type_file_extension(content_type: str) -> str:
    content_type_file_extension_mapping = {
        "application/csv": ".csv",
        "application/json": ".json",
        "application/pdf": ".pdf",
        "application/vnd.ms-excel": ".xls",
        "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet": ".xlsx",
        "application/zip": ".zip",
        "image/jpeg": ".jpeg",
        "image/png": ".png",
        "text/plain": ".txt",
        "text/xml": ".xml",
    }

    file_extension = content_type_file_extension_mapping.get(content_type, ".unknown")

    return file_extension



import requests


class Session:
    def __init__(self, headers=None, *args, **kwargs):
        self.session = requests.Session()

        if headers:
            self.session.headers.update(headers)

    def get(self, url, **kwargs):
        return self.session.get(url, **kwargs)

    def head(self, url, **kwargs):
        return self.session.head(url, **kwargs)

    def patch(self, url, **kwargs):
        return self.session.patch(url, **kwargs)

    def put(self, url, **kwargs):
        return self.session.put(url, **kwargs)

    def post(self, url, **kwargs):
        return self.session.post(url, **kwargs)

    def create(self, url, **kwargs):
        return self.post(url, **kwargs)

    def delete(self, url, **kwargs):
        return self.session.delete(url, **kwargs)



# -*- coding: utf-8 -*-
import sys
import click

from askanna.core import client as askanna_client, exceptions


class ResultGateway:
    def __init__(self, *args, **kwargs):
        self.client = askanna_client

    def get(self, run_suuid: str):
        result_url = f"{self.client.config.remote}result/{run_suuid}/"

        try:
            result = self.client.get(result_url)
        except Exception as e:
            click.echo(e, err=True)
            sys.exit(1)

        if result.status_code != 200:
            raise exceptions.GetError(
                f"{result.status_code} - We cannot find this result for you"
            )

        return result.content

    def get_content_type(self, run_suuid: str) -> str:
        result_url = f"{self.client.config.remote}result/{run_suuid}/"
        result_header = self.client.head(result_url)

        if result_header.status_code != 200:
            raise exceptions.HeadError(
                f"{result_header.status_code} - We cannot find this result for you"
            )

        return str(result_header.headers.get("Content-Type"))



"""
Management of workspaces in AskAnna
This is the class which act as gateway to the API of AskAnna
"""
import sys

from askanna.core import client, exceptions
from askanna.core.dataclasses import Workspace

import click


class WorkspaceGateway:
    def __init__(self, *args, **kwargs):
        self.client = client

    def list(self) -> list:
        # build url to select for project only
        url = "{}{}/".format(
            self.client.config.remote,
            "workspace",
        )

        r = self.client.get(url)
        if r.status_code != 200:
            raise exceptions.GetError(
                "{} - Something went wrong while retrieving "
                "workspaces: {}".format(r.status_code, r.json())
            )

        return [Workspace(**workspace) for workspace in r.json()]

    def change(
        self, suuid: str, name: str = None, description: str = None
    ) -> Workspace:
        """
        Change the workspace information
        """
        url = "{}{}/{}/".format(self.client.config.remote, "workspace", suuid)

        changes = [
            ["name", name],
            ["description", description],
        ]
        # filterout None's
        changes = list(filter(lambda x: x[1] is not None, changes))
        if len(changes) == 0:
            click.echo(
                "Nothing to change for this workspace. You did not provide a name or description.",
                err=True,
            )
            sys.exit(1)

        r = self.client.patch(url, json=dict(changes))
        if r.status_code == 200:
            workspace = Workspace(**r.json())
            click.echo(f"You have successfully changed the workspace: {workspace.name}")
            return workspace
        else:
            raise exceptions.PatchError(
                "{} - Something went wrong while updating the workspace information: "
                "{}".format(r.status_code, r.json())
            )



import os


CONFIG_FILE_ASKANNA = os.path.expanduser("~/.askanna.yml")
CONFIG_ASKANNA_REMOTE = {"askanna": {"remote": "https://beta-api.askanna.eu/v1/"}}

DEFAULT_PROJECT_TEMPLATE = "https://gitlab.askanna.io/open/project-templates/blanco-template.git"

PYPI_PROJECT_URL = "https://pypi.org/pypi/askanna/json"



# -*- coding: utf-8 -*-

"""Top-level package for askanna"""

__author__ = "AskAnna Team"
__email__ = "devops@askanna.io"
__version__ = "0.13.0"

import re
import sys

# Determine whether we are in the CLI or using the SDK
# we only check for the `askanna` executable
USING_ASKANNA_CLI: bool = any([re.match(".+bin/askanna$", sys.argv[0])])

try:
    import click  # noqa
except Exception as e:
    # We are propably within an installation
    print(e)
else:
    from .init import *  # noqa



# -*- coding: utf-8 -*-
from __future__ import absolute_import

"""Allow askanna.cli.run_utils to be executable through `python -m askanna.cli.run_utils`"""

from .tool import cli


if __name__ == "__main__":

    cli()



import os
import sys
import tempfile
import uuid
from zipfile import ZipFile

import click
import git

from askanna.core.utils import validate_askanna_yml
from askanna.core.utils import zip_files_in_dir, scan_config_in_path
from askanna.core.utils import get_config, getProjectInfo, getProjectPackages
from askanna.core.utils import extract_push_target, isIPAddress
from askanna.core.upload import PackageUpload


def package(src: str) -> str:

    pwd_dir_name = os.path.basename(src)
    random_suffix = uuid.uuid4().hex

    # make a temporary directory
    tmpdir = tempfile.mkdtemp(prefix="askanna-package")

    zip_file = os.path.join(
        tmpdir,
        "{pwd_dir_name}_{random_suffix}.zip".format(
            pwd_dir_name=pwd_dir_name, random_suffix=random_suffix
        ),
    )

    if os.path.isfile(os.path.join(src, "askannaignore")):
        ignore_file = os.path.join(src, "askannaignore")
    elif os.path.isfile(os.path.join(src, ".askannaignore")):
        ignore_file = os.path.join(src, ".askannaignore")
    elif os.path.isfile(os.path.join(src, ".gitignore")):
        ignore_file = os.path.join(src, ".gitignore")
    else:
        ignore_file = None

    cwd = os.getcwd()
    os.chdir(src)
    with ZipFile(zip_file, mode="w") as f:
        zip_files_in_dir(".", f, ignore_file)
    os.chdir(cwd)

    return zip_file


def push(force: bool, description: str = None):
    config = get_config()

    # read and parse the push-target from askanna
    push_target = config.get("push-target")

    if not push_target:
        click.echo(
            "`push-target` is not set, please set the `push-target` in the `askanna.yml` in order to push to AskAnna",
            err=True,
        )
        sys.exit(1)

    # read the config and parse jobs, validate the job definitions
    # then validate the job
    if not validate_askanna_yml(config):
        sys.exit(1)

    matches_dict = extract_push_target(push_target)
    api_host = matches_dict.get("askanna_host")
    api_server = None
    http_scheme = matches_dict.get("http_scheme")
    if api_host:
        # first also modify the first part
        if api_host.startswith("localhost") or isIPAddress(api_host.split(":")[0]):
            api_host = api_host
        elif api_host not in ["askanna.eu"]:
            # only append the -api suffix if the subdomain is not having this
            first_part = api_host.split(".")[0]
            last_part = api_host.split(".")[1:]
            if "-api" not in first_part:
                api_host = ".".join([first_part + "-api"] + last_part)
        else:
            api_host = "api." + api_host
        api_server = "{}://{}/v1/".format(http_scheme, api_host)
    project_suuid = matches_dict.get("project_suuid")

    if project_suuid:
        # make an extra call to AskAnna to query for the full uuid for this project
        project_info = getProjectInfo(project_suuid)
        if project_info.uuid is None:
            click.echo(f"Couldn't find specified project {push_target}", err=True)
            sys.exit(1)

    if not project_suuid:
        click.echo("Cannot upload to AskAnna without the project SUUID set", err=True)
        sys.exit(1)

    def ask_overwrite() -> bool:
        confirm = input("Do you want to replace the current code on AskAnna? [y/n]: ")
        answer = confirm.strip()
        if answer not in ["n", "y"]:
            print("Invalid option selected, choose from: y or n")
            return ask_overwrite()
        if confirm == "y":
            return True
        else:
            return False

    askanna_config = scan_config_in_path()
    project_folder = os.path.dirname(askanna_config)

    # check for existing package
    packages = getProjectPackages(project_info)
    if packages["count"] > 0:
        # ask for confirmation if `-f` flag is not set
        overwrite = force
        if not force:
            overwrite = ask_overwrite()

        if not overwrite:
            click.echo(
                "We are not pushing your code to AskAnna. You choose not to replace your "
                "existing code.",
                err=True,
            )
            sys.exit(0)

    package_archive = package(project_folder)

    # attach the description to this package upload
    if not description:
        # try git
        try:
            repo = git.Repo(".")
        except Exception as e:
            click.echo(e, err=True)
        else:
            commit = repo.head.commit
            description = commit.message

    # if there is still no description set then use the zipfilename
    if not description:
        description = os.path.basename(package_archive)

    click.echo("Uploading '{}' to AskAnna...".format(project_folder))

    fileinfo = {
        "filename": os.path.basename(package_archive),
        "size": os.stat(package_archive).st_size,
    }
    uploader = PackageUpload(
        api_server=api_server,
        project_suuid=project_suuid,
        description=description,
    )
    status, msg = uploader.upload(package_archive, config, fileinfo)
    if status:
        # remove temporary zip-file from the system including the parent temporary folder
        try:
            os.remove(package_archive)
            os.rmdir(os.path.dirname(package_archive))
            click.echo("Successfully pushed the project to AskAnna!")
        except OSError as e:
            click.echo(
                "Pushing your code was successful, but we could not remove the temporary file "
                "used for uploading your code to AskAnna.",
                err=True,
            )
            click.echo(f"The error: {e.strerror}", err=True)
            click.echo(f"You can manually delete the file: {package_archive}", err=True)
    else:
        click.echo(msg, err=True)
        sys.exit(1)



# -*- coding: utf-8 -*-
import os
import sys
import click
import tempfile

from askanna.cli.run_utils.utils import string_expand_variables
from askanna.core.upload import ArtifactUpload
from askanna.core.utils import get_config, create_zip_from_paths


HELP = """
Push the artifact of a run to AskAnna
"""

SHORT_HELP = "Push artifact to AskAnna"


@click.command(help=HELP, short_help=SHORT_HELP)
def cli():
    config = get_config()
    run_suuid = os.getenv("AA_RUN_SUUID")
    job_name = os.getenv("AA_JOB_NAME")

    if not run_suuid:
        click.echo("Cannot push artifact from unregistered run to AskAnna.", err=True)
        sys.exit(1)

    if not job_name:
        click.echo(
            "The job name is not set. We cannot push artifact to AskAnna.", err=True
        )
        sys.exit(1)

    # First check whether we need to create an artifact or not.
    # If output.artifact or output.paths is not specified, we skip this step and report this to the stdout.
    paths_defined = config[job_name].get("output", {}).get("artifact", [])

    if not paths_defined:  # deprecated: can be removed after release of aa-core v0.5.0
        paths_defined = config[job_name].get("output", {}).get("paths", [])
        if paths_defined:
            click.echo(
                "Deprecation warning: in a future version we remove the output/paths option. Please specify "
                "the artifact paths you want to keep in output/artifact."
            )
            click.echo(
                "More info in the documentation: https://docs.askanna.io/jobs/create-job/"
            )

    if not paths_defined:
        click.echo(
            "Artifact creation aborted, no `output/artifact` defined for this job in `askanna.yml`"
        )
        sys.exit(0)
    elif isinstance(paths_defined, str):
        paths_defined = [paths_defined]

    # Expand and translate paths if they are configured with variables
    paths = string_expand_variables(paths_defined)

    click.echo("Making a zip file with artifact...")

    tempdir = tempfile.mkdtemp(prefix="askanna-artifact")
    zip_file = os.path.join(tempdir, f"artifact_{run_suuid}.zip")
    create_zip_from_paths(filename=zip_file, paths=paths)

    click.echo("Uploading artifact to AskAnna...")

    fileinfo = {
        "filename": os.path.basename(zip_file),
        "size": os.stat(zip_file).st_size,
    }
    uploader = ArtifactUpload(
        run_suuid=run_suuid,
    )
    status, msg = uploader.upload(zip_file, config, fileinfo)
    if status:
        click.echo(msg)
        try:
            os.remove(zip_file)
            os.rmdir(tempdir)
        except OSError as e:
            click.echo(
                "Pushing your artifact was successful, but we could not remove the temporary file "
                "used for uploading the artifact to AskAnna.",
                err=True,
            )
            click.echo(f"The error: {e.strerror}", err=True)
            click.echo(f"You can manually delete the file: {zip_file}", err=True)
        sys.exit(0)
    else:
        click.echo(msg, err=True)
        sys.exit(1)



# -*- coding: utf-8 -*-
import os
import sys
import click

from askanna.core import client as askanna_client
from askanna.core.download import ChunkedDownload
from askanna.core.utils import content_type_file_extension


@click.group()
def cli1():
    pass


@cli1.command(
    help="Download a result of a run", short_help="Download a run result"
)
@click.option("--id", "-i", "suuid", prompt='Run SUUID', required=True, type=str, help="Run SUUID")
@click.option("--output", "-o", "output_path", show_default=True, type=click.Path(), help="File name to save result")
def get(suuid, output_path):
    """
    Download a result of a run
    """
    result_url = f"{askanna_client.config.remote}result/{suuid}/"
    stable_download = ChunkedDownload(result_url)

    if stable_download.status_code != 200:
        click.echo(f"{stable_download.status_code} - We cannot find this result for you", err=True)
        sys.exit(1)

    if not output_path:
        file_extension = content_type_file_extension(str(stable_download.content_type))
        output_path = f"result_{suuid}{file_extension}"

    if os.path.isdir(output_path):
        click.echo(
            "The output argument is a directory. Please provide a file name for the output.",
            err=True
        )
        sys.exit(1)

    if os.path.exists(output_path):
        click.echo(
            "The output file already exists. We will not overwrite the existing file.",
            err=True
        )
        sys.exit(1)

    click.echo("Downloading the result has started...")
    stable_download.download(output_file=output_path)
    click.echo("We have succesfully downloaded the result.")
    click.echo(f"The result is saved in: {output_path}")


cli = click.CommandCollection(
    sources=[cli1],
    help="Download result of a run",
    short_help="Download run result",
)



"""
askanna.cli specific exceptions.

Exit codes follow the sysexits.h convention:
https://www.freebsd.org/cli/man.cgi?query=sysexits&sektion=3
"""
from click import ClickException


class AskAnnaException(ClickException):
    def __init__(self, msg=None):
        super(AskAnnaException, self).__init__(msg or self.default_msg)


class AlreadyLoggedInException(AskAnnaException):
    exit_code = 0
    default_msg = ("You are already logged in. To change credentials, use "
                   "'askanna logout' first.")



import os

import click

from askanna.core.utils import store_config

HELP = """
Remove the AskAnna API key that is saved in your global configuration file
(~/.askanna.yml).
"""

SHORT_HELP = "Forget saved AskAnna API key"


@click.command(help=HELP, short_help=SHORT_HELP)
def cli():
    new_config = {"auth": {}}
    config = store_config(new_config)
    with open(os.path.expanduser("~/.askanna.yml"), "w") as fd:
        fd.write(config)
    click.echo("You have been logged out!")



import click
from cookiecutter.main import cookiecutter as cookiecreator
from cookiecutter.exceptions import OutputDirExistsException
import os
from slugify import slugify

from askanna.core import client as askanna_client
from askanna.core.utils import get_config
from askanna.core.push import push

HELP = """
This command will allow you to create an AskAnna project in a new directory
"""

SHORT_HELP = "Create a project in a new directory"


def get_user_info(server):

    url = "{server}rest-auth/user".format(server=server.replace("v1/", ""))
    ruser = askanna_client.get(url)
    if ruser.status_code == 200:
        return ruser.json()
    elif ruser.status_code == 401:
        print(
            "The provided token is not valid. Via `askanna logout` you can remove the token "
            "and via `askanna login` you can set a new token."
        )
    else:
        print("Could not connect to AskAnna at this moment")


class CreateProject:
    def __init__(self, name=None, user=None):
        self.config = get_config()
        self.api_server = self.config["askanna"]["remote"]
        self.name = name
        self.slugified_name = None
        if self.name:
            self.slugified_name = slugify(self.name)

    def find_workspace(self) -> list:
        """
        Find all workspaces where the user is part of and enumerate them
        """
        workspaces = []
        r = askanna_client.get("{}workspace/".format(self.api_server))

        for workspace in r.json():
            workspaces.append(
                {"suuid": workspace["short_uuid"], "name": workspace["name"]}
            )
        return workspaces

    def ask_workspace(self) -> str:
        """
        Ask which workspace the project will be created in.
        """
        workspaces = self.find_workspace()
        if len(workspaces) > 1:

            workspace_list_str = ""

            for idx, workspace in enumerate(workspaces, start=1):
                workspace_list_str += "%d. %s\n" % (idx, workspace["name"])

            workspace = click.prompt(
                "\nYou are a member of multiple workspaces. "
                + "In which workspace do you want to create the new project?\n"
                + workspace_list_str
                + "\n"
                + "Please enter the workspace number"
            )
            selected_workspace = workspaces[int(workspace) - 1]
        else:
            selected_workspace = workspaces[0]

        if click.confirm(
            "Do you want to create a project '{project}' in '{workspace}'?".format(
                project=self.name, workspace=selected_workspace["name"]
            ),
            abort=True,
        ):
            click.echo(
                "Thank you for the information. Anna will continue creating the project."
            )

        return selected_workspace["suuid"]

    def cli(self, workspace: str = None, description: str = None):
        if not self.name:
            click.echo(
                "Hi! It is time to create a new project in AskAnna. "
                + "We start with some information about the project."
            )
            self.name = click.prompt("Project name", type=str)
            self.slugified_name = slugify(self.name)

            if not description:
                description = click.prompt(
                    "Project description", type=str, default="", show_default=False
                )

        if not workspace:
            workspace = self.ask_workspace()

        r = askanna_client.post(
            self.api_server + "project/",
            data={
                "name": self.name,
                "workspace": workspace,
                "description": description,
            },
        )
        if r.status_code == 201:
            click.echo("\nYou have successfully created a new project in AskAnna!")
            return r.json()
        else:
            raise Exception("We could not create the project.")


@click.command(help=HELP, short_help=SHORT_HELP)
@click.argument("name", required=False)
@click.option(
    "--workspace",
    "-w",
    envvar="WORKSPACE_SUUID",
    show_default=True,
    help="Workspace SUUID where you want to create the project",
)
@click.option("--description", "-d", help="Description of the project [optional]")
@click.option(
    "--template",
    "-t",
    "project_template",
    help="Location of a Cookiecutter project template",
)
@click.option(
    "--push/--no-push",
    "-p",
    "is_push",
    default=False,
    show_default=False,
    help="Push an initial version of the code [default: no-push]",
)
def cli(name, workspace, description, project_template, is_push):
    project_creator = CreateProject(name=name)
    project_info = project_creator.cli(workspace, description)
    project_dir = project_creator.slugified_name
    if not project_template:
        project_template = project_creator.config["project"]["template"]

    try:
        cookiecreator(
            project_template,
            no_input=True,
            overwrite_if_exists=False,
            extra_context={
                "project_name": project_info["name"],
                "project_directory": project_dir,
                "project_push_target": project_info["url"],
            },
        )
    except OutputDirExistsException:
        click.echo(
            f"You already have a project directory '{project_dir}'."
            + " If you open the new project in AskAnna, you find instructions about "
            + "how you can push your project to AskAnna."
        )
    else:
        click.echo(f"Open your new local project directory: 'cd {project_dir}'")

        if is_push:
            click.echo("")  # print an empty line

            # also push the new directory to AskAnna
            os.chdir(project_dir)
            push(force=True, description="Initial push")

    # finish
    click.echo(
        "\nWe have setup the new project. You can check your project in AskAnna at:"
    )
    click.echo(project_info["url"])
    click.echo("\nSuccess with your project!")



# -*- coding: utf-8 -*-
import os
import sys
import click

from askanna.core.download import ChunkedDownload
from askanna.core.utils import get_config


@click.group()
def cli1():
    pass


@cli1.command(
    help="Download an artifact of a run", short_help="Download a run artifact"
)
@click.option("--id", "-i", "suuid", prompt='Run SUUID', required=True, type=str, help="Run SUUID")
@click.option("--output", "-o", "output_path", show_default=True, type=click.Path(), help="File name to save (zip)")
def get(suuid, output_path):
    """
    Download an artifact of a run
    """
    config = get_config()
    ASKANNA_API_SERVER = config.get("askanna", {}).get("remote")
    url = f"{ASKANNA_API_SERVER}artifact/{suuid}/"

    if not output_path:
        output_path = f"artifact_{suuid}.zip"

    if os.path.isdir(output_path):
        click.echo(
            "The output argument is a directory. Please provide a file name (zip) for the output.",
            err=True
        )
        sys.exit(1)

    if os.path.exists(output_path):
        click.echo(
            "The output file already exists. We will not overwrite the existing file.",
            err=True,
        )
        sys.exit(1)

    stable_download = ChunkedDownload(url=url)
    if stable_download.status_code != 200:
        click.echo("We cannot find this artifact for you.", err=True)
        sys.exit(1)
    click.echo("Downloading the artifact has started...")
    stable_download.download(output_file=output_path)
    click.echo("We have succesfully downloaded the artifact.")
    click.echo(f"The artifact is saved in: {output_path}")


cli = click.CommandCollection(
    sources=[cli1],
    help="Download artifact of a run",
    short_help="Download run artifact",
)



"""
Management of jobs in AskAnna
This is the class which act as gateway to the API of AskAnna
"""
import sys
import click

from askanna.core import client, exceptions
from askanna.core.dataclasses import Job


class JobGateway:
    def __init__(self, *args, **kwargs):
        self.client = client

    def list(self, project_suuid: str = None) -> list:
        if project_suuid:
            # build url to select for project only
            url = "{}{}/{}/{}/".format(
                self.client.config.remote,
                "project",
                project_suuid,
                "jobs"
            )
        else:
            url = "{}{}/".format(
                self.client.config.remote,
                "job"
            )

        r = self.client.get(url)
        if r.status_code != 200:
            raise exceptions.GetError("{} - Something went wrong while retrieving "
                                      "jobs: {}".format(r.status_code, r.json()))

        return [Job(**job) for job in r.json()]

    def get_job_by_name(self, job_name: str, project_suuid: str = None) -> Job:
        job_name = job_name.strip()
        job_list = self.list(project_suuid=project_suuid)
        result = None

        matching_jobs = list(filter(lambda x: x.name == job_name, job_list))
        if len(matching_jobs) == 0:
            raise exceptions.GetError("A job with this name is not available. Did you push your code?")

        if len(matching_jobs) > 1:
            raise exceptions.GetError("There are multiple jobs with the same name. You can narrow the selection by "
                                      "providing the project SUUID.")
        result = matching_jobs[0]
        return result

    def change(self, suuid: str, name: str = None, description: str = None) -> Job:
        url = "{}{}/{}/".format(
            self.client.config.remote,
            "job",
            suuid
        )

        job = {}
        if name:
            job.update({"name": name})
        if description:
            job.update({"description": description})

        if job:
            r = self.client.patch(url, json=job)
        else:
            click.echo("Nothing to change for this job. You did not provide a name or description.", err=True)
            sys.exit(1)

        if r.status_code == 200:
            job_updated = Job(**r.json())
            click.echo(f"You have successfully changed the job: {job_updated.name}")
        else:
            raise exceptions.PatchError("{} - Something went wrong while updating the job information: "
                                        "{}".format(r.status_code, r.json()))

        return job_updated



# -*- coding: utf-8 -*-
import importlib
import sys

from askanna import __version__ as askanna_version
import click

try:
    from dotenv import find_dotenv, load_dotenv
except ImportError:
    pass  # we only use dotenv for development
else:
    load_dotenv(find_dotenv())


HELP = """
The run util is used to support AskAnna runs
"""

SHORT_HELP = "AskAnna run util"

EPILOG = """
For usage and help on a specific command, run it with a --help flag, e.g.:

    askanna-run-utils get-payload --help
"""

CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}


def cli():
    """
    This is a wrapper to catch any command and errors
    """
    try:
        cli_commands()
    except Exception as e:
        click.echo(e, err=True)
        sys.exit(1)


@click.group(
    help=HELP, short_help=SHORT_HELP, epilog=EPILOG, context_settings=CONTEXT_SETTINGS
)
@click.version_option(version=askanna_version, prog_name="AskAnna Run Utils")
def cli_commands():
    """
    Initialize the AskAnna Run Utils commands
    """


commands = [
    "push_artifact",
    "push_metrics",
    "push_result",
    "push_variables",
    "get_package",
    "get_payload",
    "get_run_manifest",
]

for command in commands:
    module_path = "askanna.cli.run_utils." + command
    command_module = importlib.import_module(module_path)
    command_name = command.replace("_", "-")
    cli_commands.add_command(command_module.cli, command_name)



# -*- coding: utf-8 -*-
import click
import sys

from askanna import workspace as aa_workspace
from askanna.cli.utils import ask_which_workspace
from askanna.core.config import Config
from askanna.core.utils import extract_push_target


config = Config()


@click.group()
def cli1():
    pass


@cli1.command(help="List workspaces available in AskAnna", short_help="List workspaces")
def list():
    workspaces = aa_workspace.list()

    if not workspaces:
        click.echo("Based on the information provided, we cannot find any workspaces.")
        sys.exit(0)
    else:
        click.echo("WORKSPACE SUUID        WORKSPACE NAME")
        click.echo("-------------------    -------------------------")

    for workspace in sorted(workspaces, key=lambda x: (x.name,)):
        click.echo(
            "{workspace_suuid}    {workspace_name}".format(
                workspace_suuid=workspace.short_uuid,
                workspace_name=workspace.name[:25],
            )
        )


@click.group()
def cli2():
    pass


@cli2.command(
    help="Change workspace information in AskAnna", short_help="Change workspace"
)
@click.option("--id", "-i", "suuid", required=False, type=str, help="Workspace SUUID")
@click.option("--name", "-n", required=False, type=str, help="New name to set")
@click.option(
    "--description", "-d", required=False, type=str, help="New description to set"
)
def change(suuid, name, description):
    if not suuid:
        try:
            push_target = extract_push_target(config.push_target)
        except ValueError:
            # the push-target is not set, so don't bother reading it
            workspace_suuid = None
        else:
            workspace_suuid = push_target.get("workspace_suuid")

        if not workspace_suuid:
            workspace = ask_which_workspace(
                question="Which workspace do you want to change?"
            )
            workspace_suuid = workspace.short_uuid
        suuid = workspace_suuid

    if not name and not description:
        if click.confirm("\nDo you want to change the name of the workspace?"):
            name = click.prompt("New name of the workspace", type=str)
        if click.confirm("\nDo you want to change the description of the workspace?"):
            description = click.prompt("New description of the workspace", type=str)

        click.confirm("\nDo you want to change the workspace?", abort=True)

    aa_workspace.change(suuid=suuid, name=name, description=description)


cli = click.CommandCollection(
    sources=[
        cli1,
        cli2,
    ],
    help="Manage your workspaces in AskAnna",
    short_help="Manage workspaces in AskAnna",
)



"""
Management of tracking variables in AskAnna
This is the class which act as gateway to the API of AskAnna
"""
from dataclasses import dataclass
import json
import os
import sys
import tempfile
from typing import Any, List
import uuid

import click
from dateutil import parser as dateutil_parser

from askanna.core import client, exceptions
from askanna.core.dataclasses import VariableTracked, VariableDataPair, VariableLabel
from askanna.core.utils import (
    create_suuid,
    json_serializer,
    labels_to_type,
    translate_dtype,
    prepare_and_validate_value,
)

__all__ = [
    "track_variable",
    "track_variables",
    "VariableCollector",
    "VariableTrackedGateway",
]


class VariablesQuerySet:
    variables = []
    queries = []
    parent = None

    def __init__(self, variables: list = [], queries: list = [], parent=None):
        self.variables = variables
        self.queries = queries
        self.parent = parent

    def __len__(self):
        return len(self.variables)

    def __iter__(self):
        yield from self.variables

    def get(self, name, **kwargs):
        ret = self.filter(name, **kwargs)
        if len(ret) == 1:
            return ret[0]
        elif len(ret) > 1:
            click.echo(
                f"Found more variables matching name '{name}',"
                f" use .filter(name='{name}') to get all matches.",
                err=True,
            )
            return ret[0]
        else:
            click.echo(f"A variable with name '{name}' is not found")

    def filter(self, name, **kwargs):
        ret = list(
            filter(lambda x: x.get("variable", {}).get("name") == name, self.variables)
        )
        return ret

    def to_json(self) -> str:
        return json.dumps(self.variables, indent=1)

    def to_raw(self) -> list:
        return self.variables


class VariableTrackedGateway:
    def __init__(self, *args, **kwargs):
        self.client = client

    def change(self, short_uuid, variables):
        url = "{}{}/{}/{}/{}/".format(
            self.client.config.remote, "runinfo", short_uuid, "variables", short_uuid
        )

        r = self.client.patch(url, json={"variables": variables})

        if r.status_code != 200:
            raise exceptions.GetError(
                "{} - Something went wrong while updating variables "
                ": {}".format(r.status_code, r.json())
            )

    def get(self, run: str = None, job: str = None) -> VariablesQuerySet:
        """
        Depending on what we get as argument, we call the 'list' function
        with specific query params for a different endpoint
        """

        if not any([run, job]):
            # we try to get local variables if this exists

            if vc.run_suuid and vc.has_variables():
                return VariablesQuerySet(variables=vc.to_dict())
            elif vc.run_suuid:
                # RUN_SUUID is set but we don't have variables
                return VariablesQuerySet(variables=[])
            else:
                click.echo(
                    "We cannot find a run variables file for the active run.", err=True
                )
        else:
            pre_query_params = {
                "run_suuid": run,
                "job_suuid": job,
            }
            # filter out the none's:
            query_params = dict(
                filter(lambda x: x[1] is not None, pre_query_params.items())
            )
            return self.list(query_params=query_params)

    def list(self, query_params: dict = None) -> VariablesQuerySet:
        """
        List endpoint order:
        - job
        - run
        """

        endpoints = {
            "job": lambda x: "{}job/{}/variables_tracked/".format(
                self.client.config.remote, x
            ),
            "run": lambda x: "{}runinfo/{}/variables/".format(
                self.client.config.remote, x
            ),
        }
        if "run_suuid" in query_params.keys():
            url = endpoints.get("run")(query_params.get("run_suuid"))
        elif "job_suuid" in query_params.keys():
            url = endpoints.get("job")(query_params.get("job_suuid"))

        r = self.client.get(url, params=query_params)
        return VariablesQuerySet(variables=r.json())


def track_variable(name: str, value: Any, label: dict = None) -> None:
    # store the variable
    if value:
        value, valid = prepare_and_validate_value(value)
        if not valid:
            click.echo(
                f"AskAnna cannot store this datatype. Variable not stored for {name}, {value}, {label}.",
                err=True
            )
            return

    # we don't want to store secrets, so filter them and replace value if it is sensitive information
    variable_name = name.upper()
    is_masked = any(
        [
            "KEY" in variable_name,
            "TOKEN" in variable_name,
            "SECRET" in variable_name,
            "PASSWORD" in variable_name,
        ]
    )
    if is_masked:
        value = "***masked***"

    # add value to track queue
    if value:
        datapair = VariableDataPair(name=name, value=value, dtype=translate_dtype(value))
    else:
        datapair = VariableDataPair(name=name, value=None, dtype="tag")
    labels = [
        VariableLabel(name="source", value="run", dtype="string")
    ] + labels_to_type(label, labelclass=VariableLabel)

    if is_masked:
        labels.append(
            VariableLabel(**{"name": "is_masked", "value": None, "dtype": "tag"})
        )

    variable = VariableTracked(variable=datapair, label=labels)
    vc.add(variable)


def track_variables(variables: dict, label: dict = None) -> None:
    """
    Transform many variables to internal format
    """
    for name, value in variables.items():
        track_variable(name, value, label)


@dataclass
class VariableCollector:
    variables = []
    run_suuid: str = None
    local_suuid: str = None
    variables_folder: str = None
    variables_file: str = None
    changed: bool = False

    def __post_init__(self):
        if not self.run_suuid:
            # generate local suuid
            local_uuid = uuid.uuid4()
            self.local_suuid = create_suuid(local_uuid)

        self.variables_folder = self.get_variables_filepath()
        self.variables_file = os.path.join(self.variables_folder, "variables.json")

        if self.run_suuid:
            self.restore_session()

    def get_suuid(self):
        """
        Return the run_suuid if exist, otherwise the local_suuid
        """
        return self.run_suuid or self.local_suuid

    def get_variables_filepath(self):
        tmpdir = tempfile.gettempdir()
        return os.path.join(tmpdir, "askanna", "run", self.get_suuid())

    def variable_to_type(self, variable: dict = None) -> VariableDataPair:
        return VariableDataPair(
            name=variable["name"], value=variable["value"], dtype=variable["type"]
        )

    def label_to_type(self, label_list: list = None) -> list:
        _ret = []
        for label in label_list:
            _ret.append(
                VariableLabel(
                    name=label["name"], value=label["value"], dtype=label["type"]
                )
            )
        return _ret

    def restore_session(self):
        """
        We can only restore a session when the self.run_uuid is set
        """
        if not self.run_suuid:
            return

        restored_variables = []
        try:
            with open(self.variables_file, "r") as f:
                stored_variables = json.load(f)
                for stored_variable in stored_variables:
                    restored_variables.append(
                        VariableTracked(
                            variable=self.variable_to_type(
                                stored_variable.get("variable")
                            ),
                            label=self.label_to_type(stored_variable.get("label")),
                            run_suuid=stored_variable.get("run_suuid"),
                            created=dateutil_parser.isoparse(
                                stored_variable.get("created")
                            ),
                        )
                    )

                self.variables = restored_variables
                del restored_variables
        except FileNotFoundError:
            pass

    def save_session(self):
        # create folder for this session, name is the run_suuid with prefix askanna-variables

        os.makedirs(self.variables_folder, exist_ok=True)
        with open(self.variables_file, "w") as f:
            f.write(self.to_json())

    def has_variables(self) -> bool:
        return len(self.variables) > 0

    def add(self, variable: VariableTracked) -> None:
        variable.run_suuid = self.get_suuid()
        self.variables.append(variable)
        self.save_session()
        self.changed = True

    def to_dict(self, yes=True) -> List:
        return [m.to_dict() for m in self.variables]

    def to_json(self, yes=True) -> str:
        """
        We serialize all variables to json for reporting or sending to AskAnna
        """
        return json.dumps(self.to_dict(), default=json_serializer)

    def save(self, run_suuid: str = None, force: bool = False) -> None:
        """
        Connect to the AskAnna backend to save the variables
        We will only attempt to submit if there is data recorded
        """
        if not self.changed and not force:
            # We will not save when there is nothing changed (e.g. no new variable added)
            return
        if len(self.to_dict()) and not run_suuid and not self.run_suuid:
            click.echo(
                "The run SUUID is not set for this session.\n"
                + "AskAnna cannot submit the variables to the platform.",
                err=True,
            )
            click.echo(
                f"Your variables are saved locally in:\n{self.variables_file}", err=True
            )
            sys.exit(1)
        self.save_session()
        mgw = VariableTrackedGateway()
        # the run_suuid takes higher prio then the one set in init
        short_uuid = run_suuid or self.run_suuid
        mgw.change(short_uuid=short_uuid, variables=self.to_dict())


# Start logic for variable collection
vc = VariableCollector(run_suuid=os.getenv("AA_RUN_SUUID"))


# code for exiting in ipython from https://stackoverflow.com/a/40222538
# exit_register runs at the end of ipython %run or the end of the python interpreter
try:

    def exit_register(fun, *args, **kwargs):
        """Decorator that registers at post_execute. After its execution it
        unregisters itself for subsequent runs."""

        def callback():
            fun()
            ip.events.unregister("post_execute", callback)

        ip.events.register("post_execute", callback)

    ip = get_ipython()
except NameError:
    from atexit import register as exit_register


@exit_register
def handle_exit_save_variables():
    try:
        if vc.has_variables():
            vc.save()
    except Exception as e:
        # do a hard exit
        click.echo(f"Exit error in AskAnna SDK. {e}", err=True)
        os._exit(1)



# -*- coding: utf-8 -*-
import click
import sys

from askanna import job as aa_job
from askanna import project as aa_project
from askanna.cli.utils import ask_which_job, ask_which_project, ask_which_workspace
from askanna.core.config import Config
from askanna.core.utils import extract_push_target


config = Config()


@click.group()
def cli1():
    pass


@cli1.command(help="List jobs available in AskAnna",
              short_help="List jobs")
@click.option(
    "--project", "-p", "project_suuid",
    required=False,
    type=str,
    help="Project SUUID to list jobs for a project"
)
def list(project_suuid):
    jobs = aa_job.list(project_suuid)

    if not jobs:
        click.echo("Based on the information provided, we cannot find any jobs.")
        sys.exit(0)
    elif project_suuid:
        project = aa_project.detail(project_suuid)
        print("The jobs for project \"{}\" are:\n".format(project.name))
        print("JOB SUUID              JOB NAME")
        print("-------------------    -------------------------")
    else:
        print("PROJECT SUUID          PROJECT NAME            JOB SUUID              JOB NAME")
        print("-------------------    --------------------    -------------------    -------------------------")

    for job in sorted(jobs, key=lambda x: (x.project["name"], x.name)):
        if project_suuid:
            print("{job_suuid}    {job_name}".format(
                  job_suuid=job.short_uuid,
                  job_name=job.name[:25]))
        else:
            print("{project_suuid}    {project_name}    {job_suuid}    {job_name}".format(
                project_suuid=job.project["short_uuid"],
                project_name="{:20}".format(job.project["name"])[:20],
                job_suuid=job.short_uuid,
                job_name=job.name[:25]))


@click.group()
def cli2():
    pass


@cli2.command(help="Change job information in AskAnna", short_help="Change job")
@click.option("--id", "-i", "suuid", required=False, type=str, help="Job SUUID")
@click.option("--name", "-n", required=False, type=str, help="New name to set")
@click.option("--description", "-d", required=False, type=str, help="New description to set")
def change(suuid, name, description):
    if not suuid:
        try:
            push_target = extract_push_target(config.push_target)
        except ValueError as e:  # noqa
            # the push-target is not set, so don't bother reading it
            project_suuid = None
        else:
            project_suuid = push_target.get("project_suuid")

        if not project_suuid:
            workspace = ask_which_workspace(question="From which workspace do you want to change a job?")
            project = ask_which_project(question="From which project do you want to change a job?",
                                        workspace_suuid=workspace.short_uuid)
            project_suuid = project.short_uuid

        job = ask_which_job(question="Which job do you want to change?", project_suuid=project_suuid)
        suuid = job.short_uuid

    if not name and not description:
        if click.confirm("\nDo you want to change the name of the job?"):
            name = click.prompt("New name of the job", type=str)
        if click.confirm("\nDo you want to change the description of the job?"):
            description = click.prompt("New description of the job", type=str)

        click.confirm("\nDo you want to change the job?", abort=True)

    aa_job.change(suuid=suuid, name=name, description=description)


cli = click.CommandCollection(sources=[cli1, cli2], help="Manage your jobs in AskAnna",
                              short_help="Manage jobs in AskAnna")



import logging
import click
from askanna.core.variables_tracked import vc

logging.basicConfig(format="%(levelname)s:%(message)s", level=logging.INFO)

HELP = """
Retrieve variables generated with track_variable(s) and push variables to AskAnna
"""

SHORT_HELP = "Push variables to AskAnna"


@click.command(help=HELP, short_help=SHORT_HELP)
@click.option("--force", "-f", is_flag=True, help="Force push variables")
def cli(force):
    if vc.has_variables() or force:
        vc.save(force=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages
from askanna import (
    __author__ as askanna_author,
    __email__ as askanna_email,
    __version__ as askanna_version
)


with open("README.md") as readme_file:
    readme = readme_file.read()

with open("CHANGELOG.md") as history_file:
    history = history_file.read()

with open("requirements.txt") as requirements_file:
    requirements = requirements_file.read()

setup_requirements = ["wheel~=0.36.2"]


setup(
    name="askanna",
    version=askanna_version,
    author=askanna_author,
    author_email=askanna_email,
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering",
    ],
    entry_points={
        "console_scripts": [
            "askanna=askanna.cli.tool:cli",
            "askanna-run-utils=askanna.cli.run_utils.tool:cli",
        ],
    },
    description="AskAnna CLI & Python SDK is part of the AskAnna platform to kickstart your data science projects",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="askanna ml ai data",
    license="Apache License 2.0",
    packages=find_packages(exclude=["tests"]),
    install_requires=requirements,
    setup_requires=setup_requirements,
    url="https://askanna.io",
    project_urls={
        "Documentation Python SDK": "https://docs.askanna.io/python-sdk/",
        "Documentation CLI": "https://docs.askanna.io/cli/",
        "Documentation AskAnna": "https://docs.askanna.io/",
        "Release notes": "https://gitlab.com/askanna/askanna-python/-/blob/master/CHANGELOG.md",
        "Issue tracker": "https://gitlab.com/askanna/askanna-python/issues",
        "Source code": "https://gitlab.com/askanna/askanna-python",
    },
    zip_safe=False,
)



# -*- coding: utf-8 -*-
import click
import sys

from askanna import job as aa_job
from askanna import project as aa_project
from askanna import workspace as aa_workspace
from askanna.core.dataclasses import Job, Project, Workspace


def ask_which_workspace(question: str = None) -> Workspace:
    """
    Determine which workspace should be used to perform an action
    """
    workspace_list = aa_workspace.list()
    # if logged in user is only member of one workspace, we don't have to as which workspace
    if len(workspace_list) == 0:
        click.echo("It seems that you are not a member of a workspace. Please check your account.", err=True)
        sys.exit(0)
    elif len(workspace_list) == 1:
        workspace = workspace_list[0]
    else:
        list_str = ""
        for idx, workspace in enumerate(workspace_list, start=1):
            list_str += "%d. %s\n" % (idx, workspace.name)

        if not question:
            question = "Which workspace do you want to select?"

        selection = click.prompt(
            "\n" + question + "\n\n" + list_str + "\n"
            + "Please enter the workspace number",
            type=click.Choice([str(i + 1) for i in range(len(workspace_list))]),
            show_choices=False
        )

        workspace = workspace_list[int(selection) - 1]

    click.echo(f"Selected workspace: {workspace.name}")
    return workspace


def ask_which_project(question: str = None, workspace_suuid: str = None) -> Project:
    """
    Determine which project should be used to perform an action
    """
    project_list = aa_project.list(workspace_suuid=workspace_suuid)

    if len(project_list) == 0:
        click.echo("In this workspace you don't have access to any project. Please check your account or the selected "
                   "workspace.", err=True)
        sys.exit(0)

    list_str = ""
    for idx, project in enumerate(project_list, start=1):
        list_str += "%d. %s\n" % (idx, project.name)

    if not question:
        question = "Which project do you want to select?"

    selection = click.prompt(
        "\n" + question + "\n\n" + list_str + "\n"
        + "Please enter the project number",
        type=click.Choice([str(i + 1) for i in range(len(project_list))]),
        show_choices=False
    )

    project = project_list[int(selection) - 1]
    click.echo(f"Selected project: {project.name}")

    return project


def ask_which_job(question: str = None, project_suuid: str = None) -> Job:
    """
    Determine which job should be used to perform an action
    """
    job_list = aa_job.list(project_suuid=project_suuid)

    if len(job_list) == 0:
        click.echo("In this project you don't have access to any job. Please check if you have pushed your code or "
                   "check the selected project.", err=True)
        sys.exit(0)

    list_str = ""
    for idx, job in enumerate(job_list, start=1):
        list_str += "%d. %s\n" % (idx, job.name)

    if not question:
        question = "Which job do you want to select?"

    selected_job = click.prompt(
        "\n" + question + "\n\n" + list_str + "\n"
        + "Please enter the job number",
        type=click.Choice([str(i + 1) for i in range(len(job_list))]),
        show_choices=False
    )

    job = job_list[int(selected_job) - 1]
    click.echo(f"Selected job: {job.name}")

    return job



"""
Management of variables in AskAnna
This is the class which act as gateway to the API of AskAnna
"""
import sys
import click

from askanna.core import client
from askanna.core.dataclasses import Variable


class VariableGateway:
    def __init__(self, *args, **kwargs):
        self.client = client
        self.project = None
        self.cache = {}
        self.base_url = self.client.config.remote + "variable/"

    def list(self, project_suuid: str = None) -> list:
        url = self.base_url
        if project_suuid:
            # build url to select for project only
            url = "{}{}/{}/{}".format(
                self.client.config.remote,
                "project",
                project_suuid,
                "variables"
            )
        r = self.client.get(url)
        if r.status_code != 200:
            click.echo("We cannot find variables for you")
            sys.exit(1)

        variables = [Variable(**var) for var in r.json()]
        self.cache = variables
        return variables

    def detail(self, suuid: str) -> Variable:
        url = "{}{}/".format(self.base_url, suuid)
        r = self.client.get(url)
        return Variable(**r.json())

    def create(self, name: str, value: str, is_masked: bool, project_suuid: str):
        url = self.base_url
        r = self.client.create(url, json={
            "name": name,
            "value": value,
            "is_masked": is_masked,
            "project": project_suuid
        })
        if r.status_code == 201:
            return Variable(**r.json()), True

        return None, False

    def change(self, suuid: str, name: str = None, value: str = None,
               is_masked: bool = None) -> Variable:
        url = "{}{}/".format(self.base_url, suuid)

        variable = {}
        if name:
            variable.update({"name": name})
        if value:
            variable.update({"value": value})
        if is_masked:
            variable.update({"is_masked": is_masked})

        if variable:
            # emit update request (PATCH)
            r = self.client.patch(
                url,
                json=variable
            )
        else:
            click.echo("Nothing to change for this variable. You did not provide a name, value or is_masked to change "
                       "for this variable", err=True)
            sys.exit(1)

        # show success or failure
        if r.status_code == 200:
            variable_name = r.json().get('name')
            click.echo("You have successfully changed the variable: '{variable_name}'"
                       .format(variable_name=variable_name))
        else:
            click.echo("Something went wrong while changing the variable")
            sys.exit(1)
        return Variable(**r.json())

    def delete(self, suuid: str) -> bool:
        url = "{}{}/".format(self.base_url, suuid)
        r = self.client.delete(url)
        return r.status_code == 204



class Error(Exception):
    pass


class GetError(Error):
    pass


class HeadError(Error):
    pass


class PostError(Error):
    pass


class PatchError(Error):
    pass


class PutError(Error):
    pass


class RunError(Error):
    pass



from askanna.core import client, exceptions
from askanna.core.dataclasses import User
from askanna.core.utils import store_config, CONFIG_FILE_ASKANNA


class AuthGateway:
    """
    Authentication management for AskAnna CLI & SDK
    """

    def __init__(self):
        self.base_url = client.config.remote.replace("v1/", "")

    def login(
        self,
        email: str,
        password: str,
        remote: str = None,
        update_config_file: bool = False,
    ) -> str:
        if remote:
            self.base_url = remote.replace("v1/", "") or self.base_url

        url = "{remote}rest-auth/login/".format(remote=self.base_url)
        r = client.post(
            url, json={"username": email.strip(), "password": password.strip()}
        )

        if r.status_code == 400:
            raise exceptions.PostError(
                "{} - We could not log you in. Please check your credentials.".format(
                    r.status_code
                )
            )
        if r.status_code == 404:
            raise exceptions.PostError(
                "{} - We could not log you in. Please check the remote you provided.".format(
                    r.status_code
                )
            )
        elif r.status_code != 200:
            raise exceptions.PostError(
                "{} - We could not log you in: {}".format(r.status_code, r.reason)
            )

        token = str(r.json().get("key"))
        client.config.user.token = token

        if update_config_file:
            if remote:
                new_config = {"askanna": {"remote": remote}, "auth": {"token": token}}
            else:
                new_config = {"auth": {"token": token}}
            config = store_config(new_config)
            with open(CONFIG_FILE_ASKANNA, "w") as fd:
                fd.write(config)

        return token

    def get_user_info(self) -> User:
        url = "{remote}rest-auth/user".format(remote=self.base_url)

        r = client.get(url)

        if r.status_code == 200:
            return User(**r.json())
        elif r.status_code == 401:
            raise exceptions.GetError(
                "The provided token is not valid. Via `askanna logout` you can remove the token "
                "and via `askanna login` you can set a new token."
            )
        else:
            raise exceptions.GetError(
                "{} - We could not connect to AskAnna. More info:\n"
                "{}".format(r.status_code, r.reason)
            )



"""
Management of variables in AskAnna
"""
from askanna.core.dataclasses import Variable
from askanna.core.variable import VariableGateway


def list(project_suuid: str = None) -> list:
    g = VariableGateway()
    return g.list(project_suuid=project_suuid)


def detail(suuid: str) -> Variable:
    g = VariableGateway()
    return g.detail(suuid=suuid)


def create(name: str, value: str, is_masked: bool, project_suuid: str):
    g = VariableGateway()
    return g.create(name=name, value=value, is_masked=is_masked, project_suuid=project_suuid)


def change(suuid: str, name: str = None, value: str = None,
           is_masked: bool = None) -> Variable:
    g = VariableGateway()
    return g.change(suuid=suuid, name=name, value=value, is_masked=is_masked)


def delete(suuid: str) -> bool:
    g = VariableGateway()
    return g.delete(suuid=suuid)



from .base import BaseCLItest, CliRunner, cli_commands


class TestCLIJob(BaseCLItest):
    """
    askanna job

    We expect to be able to list and change jobs
    """

    verb = "job"

    def testCommandJobBase(self):
        assert "job" in self.result.output
        self.assertIn("job", self.result.output)
        self.assertNotIn("noop", self.result.output)

    def testCommadJobChange(self):
        result = CliRunner().invoke(cli_commands, "job change --help")
        assert "change [OPTIONS]" in result.output

    def testCommadJobList(self):
        result = CliRunner().invoke(cli_commands, "job list --help")
        assert "list [OPTIONS]" in result.output



import logging
import click
from askanna.core.metrics import mc

logging.basicConfig(format="%(levelname)s:%(message)s", level=logging.INFO)

HELP = """
Retrieve metrics generated with track_metric and push metrics to AskAnna
"""

SHORT_HELP = "Push metrics to AskAnna"


@click.command(help=HELP, short_help=SHORT_HELP)
@click.option("--force", "-f", is_flag=True, help="Force push metrics")
def cli(force):
    if mc.has_metrics() or force:
        mc.save(force=True)



import sys
import click

from askanna import variable as aa_variable
from askanna.cli.utils import ask_which_project, ask_which_workspace
from askanna.core.config import Config
from askanna.core.utils import extract_push_target, getProjectInfo


config = Config()

HELP = """
Manage your variables in AskAnna
"""

SHORT_HELP = "Manage variables in AskAnna"


@click.group()
def cli1():
    pass


@click.group()
def cli2():
    pass


@click.group()
def cli3():
    pass


@click.group()
def cli4():
    pass


@cli1.command(help="List variables in AskAnna", short_help="List variables")
@click.option("--project", "-p", "project_suuid", required=False, type=str,
              help="Project SUUID to list variables for a project")
def list(project_suuid):
    """
    List variables
    """
    variables = aa_variable.list(project_suuid)

    if not variables:
        click.echo("Based on the information provided, we cannot find any variables.")
        sys.exit(0)
    if project_suuid:
        click.echo("The variables for project \"{}\" are:\n".format(variables[0].project['name']))
        click.echo("VARIABLE SUUID         VARIABLE NAME")
        click.echo("-------------------    -------------------------")
    else:
        click.echo("PROJECT SUUID          PROJECT NAME            VARIABLE SUUID         VARIABLE NAME")
        click.echo("-------------------    --------------------    -------------------    -------------------------")

    for var in sorted(variables, key=lambda x: (x.project['name'], x.name)):
        if project_suuid:
            click.echo(
                "{suuid}    {variable_name}".format(
                    suuid=var.short_uuid,
                    variable_name=var.name[:25]
                )
            )
        else:
            click.echo(
                "{project_suuid}    {project_name}    {variable_suuid}    {variable_name}".format(
                    project_suuid=var.project['short_uuid'],
                    project_name="{:20}".format(var.project['name'])[:20],
                    variable_suuid=var.short_uuid,
                    variable_name=var.name[:25]
                )
            )


@cli2.command(help="Change a variable in AskAnna", short_help="Change variable")
@click.option("--id", "-i", "suuid", required=True, type=str, help="Variable SUUID")
@click.option("--name", "-n", required=False, type=str, help="New name to set")
@click.option("--value", "-v", required=False, type=str, help="New value to set")
@click.option("--masked", "-m", required=False, type=bool, help="Set value to masked")
def change(suuid, name, value, masked):
    """
    Change a variable name, value and if the value should set to be masked.
    We will only proceed when any of the name or value is set.
    """
    variable = aa_variable.detail(suuid=suuid)

    if not any([name, value]):
        click.echo("We did not change anything because you did not request to change a name or value.\n"
                   "Please add additional input to the command to change the name or value of the variable.", err=True)
        sys.exit(0)
    if masked and variable.is_masked:
        click.echo("This variable is currently masked. It is not possible to unmask a masked variable."
                   "We skip the option to change the mask status of the variable.", err=True)
        sys.exit(0)

    # commit the change of the variable to AskAnna
    aa_variable.change(suuid=suuid, name=name, value=value, is_masked=masked)


@cli3.command(help="Delete a variable in AskAnna", short_help="Delete variable")
@click.option("--id", "-i", "suuid", type=str, required=True, help="Job variable SUUID")
@click.option("--force", "-f", is_flag=True, help="Force")
def delete(suuid, force):
    """
    Delete a variable in AskAnna
    """
    try:
        variable = aa_variable.detail(suuid=suuid)
    except TypeError:
        click.echo(
            "It seems that a variable {id} doesn't exist.".format(id=id),
            err=True
        )
        sys.exit(1)

    def ask_confirmation() -> bool:
        confirm = input("Are you sure to delete variable {suuid} with name \"{name}\"? [y/n]: ".format(
            suuid=suuid, name=variable.name
        ))
        answer = confirm.strip()
        if answer not in ['n', 'y']:
            click.echo("Invalid option selected, choose from y or n")
            return ask_confirmation()
        if confirm == 'y':
            return True
        else:
            return False

    if not force:
        confirmation = ask_confirmation()
        if not confirmation:
            click.echo("Understood. We are not deleting the variable.")
            sys.exit(0)

    deleted = aa_variable.delete(suuid=suuid)
    if deleted:
        click.echo("You deleted variable {suuid}".format(suuid=suuid))
    else:
        click.echo("Something went wrong, deletion not performed.", err=True)


@cli4.command(help="Add a variable to a project in AskAnna", short_help="Add variable")
@click.option("--name", "-n", required=False, type=str, help="Name of the variable")
@click.option("--value", "-v", required=False, type=str, help="Value of the variable")
@click.option("--masked/--not-masked", "-m", "masked", default=False, show_default=False,
              help="Set value to masked (default is not-masked)")
@click.option("--project", "-p", "project_suuid", type=str, required=False,
              help="Project SUUID where this variable will be created")
def add(name, value, masked, project_suuid):
    """
    Add a variable to a project
    """

    if project_suuid:
        project = getProjectInfo(project_suuid=project_suuid)
        click.echo(f"Selected project: {project.name}")
    else:
        # Use the project from the push-target if not set
        try:
            push_target = extract_push_target(config.push_target)
        except ValueError:
            # the push-target is not set, so don't bother reading it
            pass
        else:
            project = getProjectInfo(project_suuid=project_suuid)
            if click.confirm(f"\nDo you want to create a variable for project \"{project.name}\"?"):
                project_suuid = push_target.get("project_suuid")
                click.echo(f"Selected project: {project.name}")

    if not project_suuid:
        workspace = ask_which_workspace(question="In which workspace do you want to add a variable?")
        project = ask_which_project(
            question="In which project do you want to add a variable?",
            workspace_suuid=workspace.short_uuid
        )

    confirm = False
    if not name:
        name = click.prompt("\nName of the new variable", type=str)
        confirm = True

    if not value:
        value = click.prompt("Value of the new variable", type=str)
        confirm = True

        if not masked:
            masked = click.prompt("Should the value be masked in logs and the web interface? [y/N]", type=bool)

    if confirm:
        click.confirm(f"\nDo you want to create the variable \"{name}\" in project \"{project.name}\"?", abort=True)

    variable, created = aa_variable.create(name=name, value=value, is_masked=masked, project_suuid=project.short_uuid)
    if created:
        click.echo(f"\nYou created variable \"{variable.name}\" with SUUID {variable.short_uuid} in project "
                   f"\"{project.name}\"")
        sys.exit(0)
    else:
        click.echo("Something went wrong in creating the variable.", err=True)
        sys.exit(1)


cli = click.CommandCollection(sources=[cli1, cli2, cli3, cli4], help=HELP,
                              short_help=SHORT_HELP)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

"""Allow askanna.cli to be executable through `python -m askanna.cli`"""

from .tool import cli


if __name__ == "__main__":

    cli()



import getpass
import sys

import click

from askanna.core import client as askanna_client
from askanna.core.auth import AuthGateway

HELP = """
Add your AskAnna API key to your global configuration file
(~/.askanna.yml). This is necessary to gain access to projects associated with
your AskAnna account.
"""

SHORT_HELP = "Login and save your AskAnna API key"


@click.command(help=HELP, short_help=SHORT_HELP)
@click.option('--email', '-e', required=False, type=str, help="Email which you use to login")
@click.option('--password', '-p', required=False, type=str, help="Password you use to login")
@click.option('--remote', '-r', required=False, type=str, help="Remote you want to login to")
def cli(email: str = None, password: getpass = None, remote: str = None):
    auth = AuthGateway()
    token = askanna_client.config.user.token

    if token:
        # Showcase with this token, who we are and provide info what do to if you want to use another account
        user = auth.get_user_info()
        if user.name:
            click.echo("You are already logged in as '{}' with email '{}'".format(user.name, user.email))
        else:
            click.echo("You are already logged in with email '{}'".format(user.email))

        click.echo("If you want to log in with another account, please first log out: "
                   "`askanna logout`")
        sys.exit(0)

    if email and password:
        pass
    else:
        click.echo("Let's login into AskAnna")
        if not email:
            email = click.prompt("Email", type=str)
        if not password:
            password = click.prompt("Password", type=str, hide_input=True)

    # Do the actual login and update the config file with the token
    token = auth.login(email=email, password=password, remote=remote, update_config_file=True)

    if token:
        user = auth.get_user_info()

        if user.name:
            click.echo("You are logged in as '{}' with email '{}'".format(user.name, user.email))
        else:
            click.echo("You are logged in with email '{}'".format(user.email))
